import logging
import os
from typing import List, Optional

from mir import scm

from mir.tools import mir_storage


def find_extra_items(mir_root: str) -> List[str]:
    """
    find all extra items not in mir_settings.MIR_FILES_LIST
    """
    items = os.listdir(path=mir_root)
    return list(set(items) - set(mir_storage.get_all_mir_paths()) - set(mir_storage.MIR_ASSOCIATED_FILES))


def mir_check_repo_git_dirty(mir_root: str = ".") -> bool:
    git_scm = scm.Scm(mir_root, scm_executable="git")
    git_result = git_scm.status("-s")  # if clean, returns nothing
    if (git_result or len(git_result) > 0):
        logging.info(f"git result: \n{git_result}")
        return True
    return False  # clean


def mir_check_branch_exists(mir_root: str, branch: str) -> bool:
    try:
        git_scm = scm.Scm(mir_root, scm_executable="git")
        git_scm.rev_parse(branch)
        return True
    except Exception:
        # git rev-parse will return non-zero code when can not find branch
        # and cmd.py packs non-zero return code as an error
        return False


def work_dir_to_monitor_file(work_dir: Optional[str]) -> Optional[str]:
    return os.path.join(work_dir, 'out', 'monitor.txt') if work_dir else None



import copy
from functools import wraps
import logging
import os
import shutil
import traceback
from typing import Any, Callable, Set

from mir.tools import mir_repo_utils, mir_storage_ops, phase_logger, revs_parser, utils
from mir.tools.code import MirCode
from mir.tools.errors import MirRuntimeError
from mir.protos import mir_command_pb2 as mirpb


# private: monitor.txt logger
def _get_task_name(dst_rev: str) -> str:
    return revs_parser.parse_single_arg_rev(dst_rev, need_tid=True).tid if dst_rev else 'default_task'


@utils.time_it
def _commit_error(code: int, error_msg: str, mir_root: str, src_revs: str, dst_rev: str, predefined_task: Any) -> None:
    if not src_revs:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                              error_message='empty src_revs',
                              needs_new_commit=False)
    if not dst_rev:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                              error_message='empty dst_rev',
                              needs_new_commit=False)

    src_typ_rev_tid = revs_parser.parse_arg_revs(src_revs)[0]
    dst_typ_rev_tid = revs_parser.parse_single_arg_rev(dst_rev, need_tid=True)
    if not predefined_task:
        predefined_task = mir_storage_ops.create_task(task_type=mirpb.TaskType.TaskTypeUnknown,
                                                      task_id=dst_typ_rev_tid.tid,
                                                      message='task failed',
                                                      return_code=code,
                                                      return_msg=error_msg,
                                                      src_revs=src_revs,
                                                      dst_rev=dst_rev)

    mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
                                                  mir_branch=dst_typ_rev_tid.rev,
                                                  his_branch=src_typ_rev_tid.rev,
                                                  mir_datas={},
                                                  task=predefined_task)


def _cleanup_dir_sub_items(dir: str, ignored_items: Set[str]) -> None:
    if not os.path.isdir(dir):
        return

    dir_items = os.listdir(dir)
    for item in dir_items:
        if item in ignored_items:
            continue

        item_path = os.path.join(dir, item)
        if os.path.isdir(item_path):
            shutil.rmtree(item_path)
        elif os.path.isfile(item_path):
            os.remove(item_path)


def _cleanup(work_dir: str) -> None:
    if not work_dir:
        return

    _cleanup_dir_sub_items(work_dir, ignored_items={'out'})

    _cleanup_dir_sub_items(
        os.path.join(work_dir, 'out'),
        ignored_items={
            'log.txt',  # see also: ymir-cmd-container.md
            'monitor.txt',  # monitor file
            'monitor-log.txt',  # monitor detail file
            'tensorboard',  # default root directory for tensorboard event files
            'ymir-executor-out.log',  # container output
        })


def command_run_in_out(f: Callable) -> Callable:
    """
    record monitor.txt and commit on errors
    """
    @wraps(f)
    def wrapper(mir_root: str, src_revs: str, dst_rev: str, work_dir: str, *args: tuple, **kwargs: dict) -> Any:
        mir_logger = phase_logger.PhaseLogger(task_name=_get_task_name(dst_rev),
                                              monitor_file=mir_repo_utils.work_dir_to_monitor_file(work_dir))
        mir_logger.update_percent_info(local_percent=0, task_state=phase_logger.PhaseStateEnum.PENDING)

        exc: Any = None

        try:
            ret = f(mir_root=mir_root, src_revs=src_revs, dst_rev=dst_rev, work_dir=work_dir, *args, **kwargs)
        except MirRuntimeError as e:
            error_code = e.error_code
            state_message = e.error_message
            predefined_task = e.task
            needs_new_commit = e.needs_new_commit
            exc = copy.copy(e)
            trace_message = predefined_task.return_msg if (
                predefined_task and predefined_task.return_msg) else f"cmd exception: {traceback.format_exc()}"
        except BaseException as e:
            error_code = MirCode.RC_CMD_ERROR_UNKNOWN
            state_message = str(e)
            predefined_task = None
            needs_new_commit = True
            exc = copy.copy(e)
            trace_message = f"cmd exception: {traceback.format_exc()}"
        else:
            # if no exception
            state_message = f"cmd return: {ret}"

            if ret == MirCode.RC_OK:
                mir_logger.update_percent_info(local_percent=1, task_state=phase_logger.PhaseStateEnum.DONE)
                # no need to call _commit_error, already committed inside command run function
            else:
                _commit_error(code=ret,
                              error_msg=state_message,
                              mir_root=mir_root,
                              src_revs=src_revs,
                              dst_rev=dst_rev,
                              predefined_task=None)
                mir_logger.update_percent_info(local_percent=1,
                                               task_state=phase_logger.PhaseStateEnum.ERROR,
                                               state_code=ret,
                                               state_content=state_message,
                                               trace_message='')

            logging.info(f"command done: {dst_rev}, result: {ret}")

            _cleanup(work_dir=work_dir)

            return ret

        # if MirContainerError, MirRuntimeError and BaseException occured
        # exception saved in exc
        if needs_new_commit:
            _commit_error(code=error_code,
                          error_msg=trace_message,
                          mir_root=mir_root,
                          src_revs=src_revs,
                          dst_rev=dst_rev,
                          predefined_task=predefined_task)
        mir_logger.update_percent_info(local_percent=1,
                                       task_state=phase_logger.PhaseStateEnum.ERROR,
                                       state_code=error_code,
                                       state_content=state_message,
                                       trace_message=trace_message)

        logging.info(f"command failed: {dst_rev}; exc: {exc}")
        logging.info(f"trace: {trace_message}")

        _cleanup(work_dir=work_dir)

        raise exc

    return wrapper



import argparse
import logging
import os
import subprocess
from subprocess import CalledProcessError
import traceback
from typing import Any, List, Optional, Set, Tuple

import yaml

from mir.commands import base
from mir.protos import mir_command_pb2 as mirpb
from mir.tools import checker, class_ids, context, data_exporter, mir_storage_ops, revs_parser
from mir.tools import settings as mir_settings, utils as mir_utils
from mir.tools.command_run_in_out import command_run_in_out
from mir.tools.code import MirCode
from mir.tools.errors import MirContainerError, MirRuntimeError


# private: post process
def _process_model_storage(out_root: str, model_upload_location: str, executor_config: dict,
                           task_context: dict) -> Tuple[str, float]:
    """
    find and save models
    Returns:
        model hash, model mAP and ModelStorage
    """
    out_model_dir = os.path.join(out_root, "models")
    model_paths, model_mAP = _find_models(out_model_dir)
    if not model_paths:
        # if have no models
        return '', model_mAP

    model_storage = mir_utils.ModelStorage(executor_config=executor_config,
                                           task_context=dict(**task_context,
                                                             mAP=model_mAP,
                                                             type=mirpb.TaskType.TaskTypeTraining),
                                           models=[os.path.basename(model_path) for model_path in model_paths])
    model_sha1 = mir_utils.pack_and_copy_models(model_storage=model_storage,
                                                model_dir_path=out_model_dir,
                                                model_location=model_upload_location)

    return model_sha1, model_mAP


def _find_models(model_root: str) -> Tuple[List[str], float]:
    """
    find models in `model_root`, and returns model names and mAP

    Args:
        model_root (str): model root

    Returns:
        Tuple[List[str], float]: list of model names and map
    """
    model_names = []
    model_mAP = 0.0

    result_yaml_path = os.path.join(model_root, "result.yaml")
    try:
        with open(result_yaml_path, "r") as f:
            yaml_obj = yaml.safe_load(f.read())
            model_names = yaml_obj["model"]
            model_mAP = float(yaml_obj["map"])
    except FileNotFoundError:
        logging.warning(traceback.format_exc())
        return [], 0.0

    return ([os.path.join(model_root, os.path.basename(name)) for name in model_names], model_mAP)


# private: process
def _run_train_cmd(cmd: List[str], out_log_path: str) -> int:
    """
    invoke training command

    Args:
        cmd (str): command
        out_log_path (str): path of log file

    Returns:
        int: MirCode.RC_OK if success

    Raises:
        Exception: if out_log_path can not open for append, or cmd returned non-zero code
    """
    logging.info(f"training with cmd: {cmd}")
    logging.info(f"out log path: {out_log_path}")
    with open(out_log_path, 'a') as f:
        # run and wait, if non-zero value returned, raise
        subprocess.run(cmd, check=True, stdout=f, stderr=f, text=True)

    return MirCode.RC_OK


# private: pre process
def _generate_config(executor_config: Any, out_config_path: str, task_id: str,
                     pretrained_model_params: List[str]) -> dict:
    executor_config["task_id"] = task_id
    if pretrained_model_params:
        executor_config['pretrained_model_params'] = pretrained_model_params
    elif 'pretrained_model_params' in executor_config:
        del executor_config['pretrained_model_params']

    logging.info("container config: {}".format(executor_config))

    with open(out_config_path, "w") as f:
        yaml.dump(executor_config, f)

    return executor_config


def _get_shm_size(executor_config: dict) -> str:
    if 'shm_size' not in executor_config:
        return '16G'
    return executor_config['shm_size']


def _prepare_pretrained_models(model_location: str, model_hash: str, dst_model_dir: str) -> List[str]:
    """
    prepare pretrained models
    * extract models to dst_model_dir
    * returns model file names

    Args:
        model_location (str): model location
        model_hash (str): model package hash
        dst_model_dir (str): dir where you want to extract model files to

    Returns:
        List[str]: model names
    """
    if not model_hash:
        return []
    model_storage = mir_utils.prepare_model(model_location=model_location,
                                            model_hash=model_hash,
                                            dst_model_path=dst_model_dir)

    return model_storage.models


def _get_task_parameters(config: dict) -> str:
    return config.get(mir_settings.TASK_CONTEXT_KEY, {}).get(mir_settings.TASK_CONTEXT_PARAMETERS_KEY, '')


class CmdTrain(base.BaseCommand):
    def run(self) -> int:
        logging.debug("command train: %s", self.args)

        return CmdTrain.run_with_args(work_dir=self.args.work_dir,
                                      asset_cache_dir=self.args.asset_cache_dir,
                                      model_upload_location=self.args.model_path,
                                      pretrained_model_hash=self.args.model_hash,
                                      src_revs=self.args.src_revs,
                                      dst_rev=self.args.dst_rev,
                                      mir_root=self.args.mir_root,
                                      media_location=self.args.media_location,
                                      tensorboard_dir=self.args.tensorboard_dir,
                                      executor=self.args.executor,
                                      executant_name=self.args.executant_name,
                                      config_file=self.args.config_file)

    @staticmethod
    @command_run_in_out
    def run_with_args(work_dir: str,
                      asset_cache_dir: Optional[str],
                      model_upload_location: str,
                      pretrained_model_hash: str,
                      executor: str,
                      executant_name: str,
                      src_revs: str,
                      dst_rev: str,
                      config_file: Optional[str],
                      tensorboard_dir: str,
                      mir_root: str = '.',
                      media_location: str = '') -> int:
        if not model_upload_location:
            logging.error("empty --model-location, abort")
            return MirCode.RC_CMD_INVALID_ARGS
        src_typ_rev_tid = revs_parser.parse_single_arg_rev(src_revs, need_tid=False)
        dst_typ_rev_tid = revs_parser.parse_single_arg_rev(dst_rev, need_tid=True)
        if not work_dir:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='empty work_dir')
        if not config_file or not os.path.isfile(config_file):
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                                  error_message=f"invalid --task-config-file: {config_file}")
        if asset_cache_dir and not os.path.isabs(asset_cache_dir):
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                                  error_message=f"invalid --cache {config_file}, not an absolute path for directory")

        return_code = checker.check(mir_root,
                                    [checker.Prerequisites.IS_INSIDE_MIR_REPO, checker.Prerequisites.HAVE_LABELS])
        if return_code != MirCode.RC_OK:
            return return_code

        with open(config_file, "r") as f:
            config = yaml.safe_load(f)

        task_parameters = _get_task_parameters(config)
        if not isinstance(task_parameters, str):
            raise MirRuntimeError(
                error_code=MirCode.RC_CMD_INVALID_ARGS,
                error_message=f"invalid {mir_settings.TASK_CONTEXT_PARAMETERS_KEY} in config: {config}")
        if mir_settings.EXECUTOR_CONFIG_KEY not in config:
            raise MirRuntimeError(
                error_code=MirCode.RC_CMD_INVALID_ARGS,
                error_message=f"invalid config file: {config_file}, needs: {mir_settings.EXECUTOR_CONFIG_KEY}")

        executor_config = config[mir_settings.EXECUTOR_CONFIG_KEY]

        class_names = executor_config.get('class_names', [])
        if not class_names:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                                  error_message=f"no class_names in config file: {config_file}")
        if len(set(class_names)) != len(class_names):
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                                  error_message=f"dumplicate class names in class_names: {class_names}")

        task_id = dst_typ_rev_tid.tid
        if not executant_name:
            executant_name = f"default-training-{task_id}"
        if not tensorboard_dir:
            tensorboard_dir = os.path.join(work_dir, 'out', 'tensorboard')
        asset_dir = asset_cache_dir or os.path.join(work_dir, 'in', 'assets')

        # if have model_hash, export model
        pretrained_model_names = _prepare_pretrained_models(model_location=model_upload_location,
                                                            model_hash=pretrained_model_hash,
                                                            dst_model_dir=os.path.join(work_dir, 'in', 'models'))

        # get train_ids and val_ids
        train_ids = set()  # type: Set[str]
        val_ids = set()  # type: Set[str]
        unused_ids = set()  # type: Set[str]
        mir_metadatas: mirpb.MirMetadatas = mir_storage_ops.MirStorageOps.load_single_storage(
            mir_root=mir_root,
            mir_branch=src_typ_rev_tid.rev,
            mir_task_id=src_typ_rev_tid.tid,
            ms=mirpb.MirStorage.MIR_METADATAS)
        for asset_id, asset_attr in mir_metadatas.attributes.items():
            if asset_attr.tvt_type == mirpb.TvtTypeTraining:
                train_ids.add(asset_id)
            elif asset_attr.tvt_type == mirpb.TvtTypeValidation:
                val_ids.add(asset_id)
            else:
                unused_ids.add(asset_id)
        if not train_ids:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='no training set')
        if not val_ids:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='no validation set')

        if not unused_ids:
            logging.info(f"training: {len(train_ids)}, validation: {len(val_ids)}")
        else:
            logging.warning(f"training: {len(train_ids)}, validation: {len(val_ids)}" f"unused: {len(unused_ids)}")

        # export
        logging.info("exporting assets")

        os.makedirs(work_dir, exist_ok=True)

        work_dir_in = os.path.join(work_dir, "in")
        work_dir_annotations = os.path.join(work_dir_in, 'annotations')
        os.makedirs(work_dir_annotations, exist_ok=True)

        work_dir_out = os.path.join(work_dir, "out")
        os.makedirs(work_dir_out, exist_ok=True)
        out_model_dir = os.path.join(work_dir, 'out', 'models')
        os.makedirs(out_model_dir, exist_ok=True)

        os.makedirs(asset_dir, exist_ok=True)
        os.makedirs(tensorboard_dir, exist_ok=True)

        # type names to type ids
        # ['cat', 'person'] -> [4, 2]
        cls_mgr = class_ids.ClassIdManager(mir_root=mir_root)
        type_ids_list = cls_mgr.id_for_names(class_names)
        if not type_ids_list:
            logging.info(f"type ids empty, please check config file: {config_file}")
            return MirCode.RC_CMD_INVALID_ARGS

        if not context.check_class_ids(mir_root=mir_root, current_class_ids=type_ids_list):
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='user class ids mismatch')

        type_id_idx_mapping = {type_id: index for (index, type_id) in enumerate(type_ids_list)}

        # export train set
        data_exporter.export(mir_root=mir_root,
                             assets_location=media_location,
                             class_type_ids=type_id_idx_mapping,
                             asset_ids=train_ids,
                             asset_dir=asset_dir,
                             annotation_dir=work_dir_annotations,
                             need_ext=True,
                             need_id_sub_folder=True,
                             base_branch=src_typ_rev_tid.rev,
                             base_task_id=src_typ_rev_tid.tid,
                             format_type=data_exporter.ExportFormat.EXPORT_FORMAT_ARK,
                             index_file_path=os.path.join(work_dir_in, 'train-index.tsv'),
                             index_assets_prefix='/in/assets',
                             index_annotations_prefix='/in/annotations')

        # export validation set
        data_exporter.export(mir_root=mir_root,
                             assets_location=media_location,
                             class_type_ids=type_id_idx_mapping,
                             asset_ids=val_ids,
                             asset_dir=asset_dir,
                             annotation_dir=work_dir_annotations,
                             need_ext=True,
                             need_id_sub_folder=True,
                             base_branch=src_typ_rev_tid.rev,
                             base_task_id=src_typ_rev_tid.tid,
                             format_type=data_exporter.ExportFormat.EXPORT_FORMAT_ARK,
                             index_file_path=os.path.join(work_dir_in, 'val-index.tsv'),
                             index_assets_prefix='/in/assets',
                             index_annotations_prefix='/in/annotations')

        logging.info("starting train docker container")

        available_gpu_id = config.get(mir_settings.TASK_CONTEXT_KEY, {}).get('available_gpu_id', '')

        # generate configs
        out_config_path = os.path.join(work_dir_in, "config.yaml")
        executor_config = _generate_config(
            executor_config=executor_config,
            out_config_path=out_config_path,
            task_id=task_id,
            pretrained_model_params=[os.path.join('/in/models', name) for name in pretrained_model_names])
        mir_utils.generate_training_env_config_file(task_id=task_id,
                                                    env_config_file_path=os.path.join(work_dir_in, 'env.yaml'))

        # start train docker and wait
        path_binds = []
        path_binds.append(f"-v{work_dir_in}:/in")  # annotations, models, train-index.tsv, val-index.tsv, config.yaml
        path_binds.append(f"-v{asset_dir}:/in/assets:ro")  # assets
        path_binds.append(f"-v{work_dir_out}:/out")
        path_binds.append(f"-v{tensorboard_dir}:/out/tensorboard")

        cmd = ['nvidia-docker', 'run', '--rm', f"--shm-size={_get_shm_size(executor_config=executor_config)}"]
        cmd.extend(path_binds)
        if available_gpu_id:
            cmd.extend(['--gpus', f"\"device={available_gpu_id}\""])
        cmd.extend(['--user', f"{os.getuid()}:{os.getgid()}"])  # run as current user
        cmd.extend(['--name', f"{executant_name}"])  # executor name used to stop executor
        cmd.append(executor)

        task_code = MirCode.RC_OK
        return_msg = ''
        try:
            _run_train_cmd(cmd, out_log_path=os.path.join(work_dir_out, mir_settings.EXECUTOR_OUTLOG_NAME))
        except CalledProcessError as e:
            logging.warning(f"training exception: {e}")
            # don't exit, proceed if model exists
            task_code = MirCode.RC_CMD_CONTAINER_ERROR
            return_msg = mir_utils.collect_executor_outlog_tail(work_dir=work_dir)

        # gen task_context
        task_context = {
            'src_revs': src_revs,
            'dst_rev': dst_rev,
            'executor': executor,
            mir_settings.PRODUCER_KEY: mir_settings.PRODUCER_NAME,
            mir_settings.TASK_CONTEXT_PARAMETERS_KEY: task_parameters
        }

        # save model
        logging.info("saving models")
        model_sha1, model_mAP = _process_model_storage(out_root=work_dir_out,
                                                       model_upload_location=model_upload_location,
                                                       executor_config=executor_config,
                                                       task_context=task_context)

        # commit task
        task = mir_storage_ops.create_task(task_type=mirpb.TaskType.TaskTypeTraining,
                                           task_id=dst_typ_rev_tid.tid,
                                           message='training',
                                           model_mAP=model_mAP,
                                           model_hash=model_sha1,
                                           return_code=task_code,
                                           return_msg=return_msg,
                                           serialized_task_parameters=task_parameters,
                                           serialized_executor_config=yaml.safe_dump(executor_config),
                                           executor=executor,
                                           src_revs=src_revs,
                                           dst_rev=dst_rev)

        if task_code != MirCode.RC_OK:
            raise MirContainerError(error_message='container error occured', task=task)

        mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
                                                      mir_branch=dst_typ_rev_tid.rev,
                                                      his_branch=src_typ_rev_tid.rev,
                                                      mir_datas={},
                                                      task=task)

        logging.info("training done")

        return MirCode.RC_OK


def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: argparse.ArgumentParser) -> None:
    train_arg_parser = subparsers.add_parser("train",
                                             parents=[parent_parser],
                                             description="use this command to train current workspace",
                                             help="train current workspace")
    train_arg_parser.add_argument("--model-location",
                                  required=True,
                                  dest="model_path",
                                  type=str,
                                  help="storage place (upload location) to store packed model")
    train_arg_parser.add_argument("--media-location",
                                  required=True,
                                  dest="media_location",
                                  type=str,
                                  help="media storage location for models")
    train_arg_parser.add_argument('--model-hash',
                                  dest='model_hash',
                                  type=str,
                                  required=False,
                                  help='model hash to be used')
    train_arg_parser.add_argument("-w", required=True, dest="work_dir", type=str, help="work place for training")
    train_arg_parser.add_argument('--asset-cache-dir',
                                  required=False,
                                  dest='asset_cache_dir',
                                  type=str,
                                  help='asset cache directory')
    train_arg_parser.add_argument("--executor",
                                  required=True,
                                  dest="executor",
                                  type=str,
                                  help="docker image name for training")
    train_arg_parser.add_argument('--executant-name',
                                  required=False,
                                  dest='executant_name',
                                  type=str,
                                  help='docker container name for training')
    train_arg_parser.add_argument("--src-revs",
                                  dest="src_revs",
                                  type=str,
                                  required=True,
                                  help="rev@bid: source rev and base task id")
    train_arg_parser.add_argument("--dst-rev",
                                  dest="dst_rev",
                                  type=str,
                                  required=True,
                                  help="rev@tid: destination branch name and task id")
    train_arg_parser.add_argument("--task-config-file",
                                  dest="config_file",
                                  type=str,
                                  required=True,
                                  help="path to executor config file")
    train_arg_parser.add_argument("--tensorboard-dir",
                                  dest="tensorboard_dir",
                                  type=str,
                                  required=False,
                                  help="tensorboard log directory")
    train_arg_parser.set_defaults(func=CmdTrain)



import os

from mir.scm.cmd import CmdScm
from mir.tools.code import MirCode
from mir.tools.errors import MirRuntimeError


def Scm(root_dir: str, scm_executable: str = None) -> CmdScm:
    """Returns SCM instance that corresponds to a repo at the specified
        path.
        Args:
            root_dir (str): path to a root directory of the repo.
            scm_excutable(str): "git".
        Returns:
            mir.scm.cmd.BaseScm: SCM instance.
        """

    if scm_executable not in ["git"]:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                              error_message=f"args error: expected git, not {scm_executable}")
    if not os.path.exists(root_dir):
        os.makedirs(root_dir)
    if not os.path.isdir(root_dir):
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                              error_message=f"can not create dir: {root_dir}")
    return CmdScm(root_dir, scm_executable)



import argparse
import logging
import random

from mir.commands import base
from mir.protos import mir_command_pb2 as mirpb
from mir.tools import mir_storage_ops, revs_parser
from mir.tools.code import MirCode
from mir.tools.command_run_in_out import command_run_in_out
from mir.tools.errors import MirRuntimeError


class CmdSampling(base.BaseCommand):
    def run(self) -> int:
        logging.debug("command sampling: %s", self.args)
        return CmdSampling.run_with_args(mir_root=self.args.mir_root,
                                         work_dir=self.args.work_dir,
                                         src_revs=self.args.src_revs,
                                         dst_rev=self.args.dst_rev,
                                         count=self.args.count,
                                         rate=self.args.rate)

    @staticmethod
    @command_run_in_out
    def run_with_args(mir_root: str, work_dir: str, src_revs: str, dst_rev: str, count: int, rate: float) -> int:
        src_typ_rev_tid = revs_parser.parse_single_arg_rev(src_revs, need_tid=False)
        dst_typ_rev_tid = revs_parser.parse_single_arg_rev(dst_rev, need_tid=True)

        if count < 0:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='invalid args: --count')
        if rate < 0 or rate > 1:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='invalid args: --rate')

        if not mir_root:
            mir_root = '.'

        # read all
        [mir_metadatas, mir_annotations, mir_tasks] = mir_storage_ops.MirStorageOps.load_multiple_storages(
            mir_root=mir_root,
            mir_branch=src_typ_rev_tid.rev,
            mir_task_id=src_typ_rev_tid.tid,
            ms_list=[mirpb.MirStorage.MIR_METADATAS, mirpb.MirStorage.MIR_ANNOTATIONS, mirpb.MirStorage.MIR_TASKS],
            as_dict=False,
        )
        assets_count = len(mir_metadatas.attributes)
        sampled_assets_count = 0
        if count > 0:
            sampled_assets_count = count
        else:
            sampled_assets_count = int(assets_count * rate)
        if sampled_assets_count < 0:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                                  error_message=f"sampled assets count: {sampled_assets_count} is negative")
        if sampled_assets_count > assets_count:
            logging.warning(f"sampled assets count: {sampled_assets_count} > assets count: {assets_count}, select all")
            sampled_assets_count = assets_count

        # sampling
        if sampled_assets_count < assets_count:
            sampled_asset_ids = random.sample(mir_metadatas.attributes.keys(), sampled_assets_count)
            # sampled_mir_metadatas and sampled_mir_annotations
            image_annotations = mir_annotations.task_annotations[mir_annotations.head_task_id].image_annotations
            sampled_mir_metadatas = mirpb.MirMetadatas()
            sampled_mir_annotations = mirpb.MirAnnotations()
            for asset_id in sampled_asset_ids:
                sampled_mir_metadatas.attributes[asset_id].CopyFrom(mir_metadatas.attributes[asset_id])
                if asset_id in image_annotations:
                    sampled_mir_annotations.task_annotations[dst_typ_rev_tid.tid].image_annotations[asset_id].CopyFrom(
                        image_annotations[asset_id])
        else:
            # if equals
            sampled_mir_metadatas = mir_metadatas
            sampled_mir_annotations = mirpb.MirAnnotations()
            sampled_mir_annotations.head_task_id = dst_typ_rev_tid.tid
            sampled_mir_annotations.task_annotations[dst_typ_rev_tid.tid].CopyFrom(
                mir_annotations.task_annotations[mir_annotations.head_task_id])

        # mir_tasks
        message = f"sampling src: {src_revs}, dst: {dst_rev}, count: {count}, rate: {rate}"
        task = mir_storage_ops.create_task(task_type=mirpb.TaskType.TaskTypeSampling,
                                           task_id=dst_typ_rev_tid.tid,
                                           message=message,
                                           src_revs=src_revs,
                                           dst_rev=dst_rev)

        logging.info(f"sampling done, assets count: {sampled_assets_count}")

        # save and commit
        sampled_mir_datas = {
            mirpb.MirStorage.MIR_METADATAS: sampled_mir_metadatas,
            mirpb.MirStorage.MIR_ANNOTATIONS: sampled_mir_annotations,
        }
        mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
                                                      mir_branch=dst_typ_rev_tid.rev,
                                                      his_branch=src_typ_rev_tid.rev,
                                                      mir_datas=sampled_mir_datas,
                                                      task=task)

        return MirCode.RC_OK


def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: argparse.ArgumentParser) -> None:
    sampling_arg_parser = subparsers.add_parser('sampling',
                                                parents=[parent_parser],
                                                description='use this command to sample assets',
                                                help='sample assets')
    sampling_arg_parser.add_argument('--src-revs', dest='src_revs', type=str, help='rev@bid')
    sampling_arg_parser.add_argument('--dst-rev', dest='dst_rev', type=str, help='rev@tid')
    sampling_arg_parser.add_argument('-w', dest='work_dir', type=str, required=False, help='working directory')
    group = sampling_arg_parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--count', dest='count', type=int, default=0, help='assets count')
    group.add_argument('--rate', dest='rate', type=float, default=0.0, help='assets sampling rate')
    sampling_arg_parser.set_defaults(func=CmdSampling)



""" write task progress percent """

import datetime
from enum import IntEnum
import json
import math
import os
from typing import Any, Dict, List, Optional

from mir.protos import mir_command_pb2 as mirpb
from mir.tools.code import MirCode
from mir.tools.errors import MirError


class PhaseStateEnum(IntEnum):
    PENDING = mirpb.TaskStatePending
    RUNNING = mirpb.TaskStateRunning
    DONE = mirpb.TaskStateDone
    ERROR = mirpb.TaskStateError


class PhaseLoggerError(MirError):
    def __init__(self, error_message: str) -> None:
        super().__init__(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message=error_message)


def _raise_if_false(condition: Any, msg: str) -> None:
    if not condition:
        raise PhaseLoggerError(error_message=msg)


class PhaseLogger:
    # life cycle
    def __init__(self,
                 task_name: str,
                 monitor_file: Optional[str] = None,
                 start: float = 0.0,
                 end: float = 1.0,
                 auto_done: bool = False) -> None:
        _raise_if_false(task_name, 'empty task name')
        _raise_if_false(start >= 0 and start <= 1, 'invalid start')
        _raise_if_false(end >= 0 and end <= 1, 'invalid end')
        _raise_if_false(start < end, 'start ge end')

        self._task_name = task_name
        self._monitor_file = monitor_file
        self._start_percent = start
        self._end_percent = end
        self._local_percent = 0.0  # between 0 and 1
        self.auto_done = auto_done

        if monitor_file:
            monitor_dir_name = os.path.dirname(monitor_file)
            if not os.path.isdir(monitor_dir_name):
                os.makedirs(monitor_dir_name, exist_ok=True)

    # public: properties
    @property
    def task_name(self) -> str:
        return self._task_name

    @property
    def monitor_file(self) -> Optional[str]:
        return self._monitor_file

    @property
    def start_percent(self) -> float:
        return self._start_percent

    @property
    def end_percent(self) -> float:
        return self._end_percent

    @property
    def global_percent(self) -> float:
        return self._start_percent + self._local_percent * (self._end_percent - self._start_percent)

    @property
    def local_percent(self) -> float:
        return self._local_percent

    # public: general
    def update_percent_info(self, local_percent: float, task_state: PhaseStateEnum,
                            state_code: int = 0,
                            state_content: str = None,
                            trace_message: str = None) -> None:
        # if no monitor_file assgned, no need to write monitor percent log
        _raise_if_false(local_percent >= 0 and local_percent <= 1, 'invalid local_percent')
        self._local_percent = local_percent

        if not self.monitor_file:
            return

        global_percent = min(self.global_percent, 1)

        with open(self.monitor_file, 'w') as f:
            timestamp = datetime.datetime.now().timestamp()
            f.write(f"{self.task_name}\t{timestamp:.6f}\t{global_percent:.2f}\t{task_state}")
            if state_code and state_content:
                f.write(f"\t{state_code}\t{state_content}")
            f.write("\n")
            if trace_message:
                f.write(f"{trace_message}\n")

    def create_children(self, deltas: List[float]) -> List['PhaseLogger']:
        """
        creates child phase loggers for current period

        for example:
        * a phase logger in period (start=0, end=0) will separate into two sub loggers for delta [0.4, 0.6]:
            * sub logger 1: start=0, end=0.4
            * sub logger 2: start=0.4, end=1
        * a phase logger in period (start=0.1, end=0.8) will separate into two sub loggers for delta [0.4, 0.6]:
            * sub logger 1: start=0.1, end=0.38
            * sub logger 2: start=0.38, end=0.8

        Args:
            deltas (List[float]): separation ratios

        Returns:
            List[PhaseLogger]: child phase loggers
        """
        # check deltas
        total = 0.0
        for d in deltas:
            _raise_if_false(d > 0 and d <= 1, 'invalid delta')
            total += d
        _raise_if_false(math.isclose(total, 1), f"total: {total}")

        start = self.start_percent
        parent_delta = self.end_percent - self.start_percent
        children: List[PhaseLogger] = []
        for d in deltas:
            end = start + d * parent_delta
            children.append(PhaseLogger(task_name=self.task_name, monitor_file=self.monitor_file, start=start, end=end))
            start = end

        return children


class PhaseLoggerCenter:
    _phase_name_to_loggers: Dict[str, PhaseLogger] = {}

    @staticmethod
    def clear_all() -> None:
        if PhaseLoggerCenter._phase_name_to_loggers:
            PhaseLoggerCenter._phase_name_to_loggers = {}

    @staticmethod
    def loggers() -> Dict[str, PhaseLogger]:
        return dict(PhaseLoggerCenter._phase_name_to_loggers)

    @staticmethod
    def create_phase_loggers(top_phase: str, monitor_file: Optional[str], task_name: str) -> None:
        conf_path = os.path.join(os.path.dirname(__file__), 'phase_logger_conf.json')
        with open(conf_path, 'r') as f:
            conf = json.loads(f.read())
        _raise_if_false(top_phase in conf, f"{top_phase} not in {conf}")

        sub_phase_confs = conf[top_phase]
        sub_phase_names = []
        sub_deltas: List[Any] = []
        for sub_phase_conf in sub_phase_confs:
            sub_phase_names.append(sub_phase_conf['name'])
            sub_deltas.append(sub_phase_conf['delta'])

        sub_phase_count = len(sub_phase_names)
        top_logger = PhaseLogger(task_name=task_name, monitor_file=monitor_file)
        sub_phase_loggers = top_logger.create_children(deltas=sub_deltas)
        for idx, (sub_phase_name, sub_phase_logger) in enumerate(zip(sub_phase_names, sub_phase_loggers)):
            if idx == sub_phase_count - 1:  # if last, let whole phase auto done
                sub_phase_logger.auto_done = True
            PhaseLoggerCenter._phase_name_to_loggers[f"{top_phase}.{sub_phase_name}"] = sub_phase_logger

    @staticmethod
    def update_phase(phase: str,
                     local_percent: float = 1,
                     task_state: PhaseStateEnum = PhaseStateEnum.RUNNING,
                     state_code: int = 0,
                     state_content: str = None,
                     trace_message: str = None) -> None:
        if not phase:
            return

        _raise_if_false(phase in PhaseLoggerCenter._phase_name_to_loggers, f"unknown phase: {phase}")

        PhaseLoggerCenter._phase_name_to_loggers[phase].update_percent_info(local_percent=local_percent,
                                                                            task_state=task_state,
                                                                            state_code=state_code,
                                                                            state_content=state_content,
                                                                            trace_message=trace_message)



import argparse
import logging
import json
import os
import random
import shutil

from mir.commands import base
from mir.protos import mir_command_pb2 as mirpb
from mir.tools import annotations, checker, hash_utils, metadatas, mir_repo_utils, mir_storage_ops, revs_parser
from mir.tools.code import MirCode
from mir.tools.command_run_in_out import command_run_in_out
from mir.tools.errors import MirRuntimeError
from mir.tools.phase_logger import PhaseLoggerCenter


class CmdImport(base.BaseCommand):
    def run(self) -> int:
        logging.debug("command import: %s", self.args)

        return CmdImport.run_with_args(mir_root=self.args.mir_root,
                                       index_file=self.args.index_file,
                                       ck_file='',
                                       anno_abs=self.args.anno,
                                       gen_abs=self.args.gen,
                                       dataset_name=self.args.dataset_name,
                                       dst_rev=self.args.dst_rev,
                                       src_revs=self.args.src_revs or 'master',
                                       work_dir=self.args.work_dir,
                                       ignore_unknown_types=self.args.ignore_unknown_types)

    @staticmethod
    @command_run_in_out
    def run_with_args(mir_root: str, index_file: str, ck_file: str, anno_abs: str, gen_abs: str, dataset_name: str,
                      dst_rev: str, src_revs: str, work_dir: str, ignore_unknown_types: bool) -> int:
        # Step 1: check args and prepare environment.
        if not index_file or not gen_abs or not os.path.isfile(index_file):
            logging.error(f"invalid index_file: {index_file} or gen_abs: {gen_abs}")
            return MirCode.RC_CMD_INVALID_ARGS
        if anno_abs and not os.path.isdir(anno_abs):
            logging.error(f"annotations dir invalid: {anno_abs}")
            return MirCode.RC_CMD_INVALID_ARGS
        dst_typ_rev_tid = revs_parser.parse_single_arg_rev(dst_rev, need_tid=True)
        src_typ_rev_tid = revs_parser.parse_single_arg_rev(src_revs, need_tid=False)

        if not dataset_name:
            dataset_name = dst_typ_rev_tid.tid

        PhaseLoggerCenter.create_phase_loggers(top_phase='import',
                                               monitor_file=mir_repo_utils.work_dir_to_monitor_file(work_dir),
                                               task_name=dst_typ_rev_tid.tid)

        check_code = checker.check(mir_root,
                                   [checker.Prerequisites.IS_INSIDE_MIR_REPO, checker.Prerequisites.HAVE_LABELS])
        if check_code != MirCode.RC_OK:
            return check_code

        # Step 2: generate sha1 file and rename images.
        # sha1 file to be written.
        sha1_index_abs = os.path.join(
            gen_abs, f"{os.path.basename(index_file)}-{dst_typ_rev_tid.rev}-{random.randint(0, 100)}.sha1")
        ret = _generate_sha_and_copy(index_file, sha1_index_abs, gen_abs)
        if ret != MirCode.RC_OK:
            logging.error(f"generate hash error: {ret}")
            return ret

        # Step 3 import metadat and annotations:
        mir_metadatas = mirpb.MirMetadatas()
        ret = metadatas.import_metadatas(mir_metadatas=mir_metadatas,
                                         dataset_name=dataset_name,
                                         in_sha1_path=sha1_index_abs,
                                         hashed_asset_root=gen_abs,
                                         phase='import.metadatas')
        if ret != MirCode.RC_OK:
            logging.error(f"import metadatas error: {ret}")
            return ret

        mir_annotation = mirpb.MirAnnotations()
        mir_keywords = mirpb.MirKeywords()
        ret_code, unknown_types = annotations.import_annotations(mir_annotation=mir_annotation,
                                                                 mir_keywords=mir_keywords,
                                                                 in_sha1_file=sha1_index_abs,
                                                                 ck_file=ck_file,
                                                                 mir_root=mir_root,
                                                                 annotations_dir_path=anno_abs,
                                                                 task_id=dst_typ_rev_tid.tid,
                                                                 phase='import.others')
        if ret_code != MirCode.RC_OK:
            logging.error(f"import annotations error: {ret_code}")
            return ret_code
        if unknown_types:
            if ignore_unknown_types:
                logging.warning(f"unknown types: {unknown_types}")
            else:
                raise MirRuntimeError(MirCode.RC_CMD_UNKNOWN_TYPES, json.dumps(unknown_types))

        # create and write tasks
        task = mir_storage_ops.create_task(task_type=mirpb.TaskTypeImportData,
                                           task_id=dst_typ_rev_tid.tid,
                                           message=f"importing {index_file}-{anno_abs}-{gen_abs} as {dataset_name}",
                                           unknown_types=unknown_types,
                                           src_revs=src_revs,
                                           dst_rev=dst_rev)

        mir_data = {
            mirpb.MirStorage.MIR_METADATAS: mir_metadatas,
            mirpb.MirStorage.MIR_ANNOTATIONS: mir_annotation,
        }
        mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
                                                      his_branch=src_typ_rev_tid.rev,
                                                      mir_branch=dst_typ_rev_tid.rev,
                                                      mir_datas=mir_data,
                                                      task=task)

        # cleanup
        os.remove(sha1_index_abs)

        return MirCode.RC_OK


def _generate_sha_and_copy(index_file: str, sha_idx_file: str, sha_folder: str) -> int:
    hash_phase_name = 'import.hash'
    if not os.path.isfile(index_file):
        logging.error('invalid index_file')
        return MirCode.RC_CMD_INVALID_ARGS

    os.makedirs(sha_folder, exist_ok=True)

    with open(index_file) as idx_f, open(sha_idx_file, 'w') as sha_f:
        lines = idx_f.readlines()
        total_count = len(lines)
        asset_count_limit = 1000000
        if total_count > asset_count_limit:  # large number of images may trigger redis timeout error.
            logging.error(f'# of image {total_count} exceeds upper boundary {asset_count_limit}.')
            return MirCode.RC_CMD_INVALID_ARGS

        idx = 0
        for line in lines:
            media_src = line.strip()
            if not media_src or not os.path.isfile(media_src):
                logging.warning("invalid file: ", media_src)
                continue
            sha1 = hash_utils.sha1sum_for_file(media_src)
            sha_f.writelines("\t".join([sha1, media_src]) + '\n')

            media_dst = os.path.join(sha_folder, sha1)
            if not os.path.isfile(media_dst):
                shutil.copyfile(media_src, media_dst)

            idx += 1
            if idx % 5000 == 0:
                PhaseLoggerCenter.update_phase(phase=hash_phase_name, local_percent=(idx / total_count))
                logging.info(f"finished {idx} / {total_count} hashes")
    PhaseLoggerCenter.update_phase(phase=hash_phase_name)
    return MirCode.RC_OK


def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: argparse.ArgumentParser) -> None:
    importing_arg_parser = subparsers.add_parser("import",
                                                 parents=[parent_parser],
                                                 description="use this command to import data from img/anno folder",
                                                 help="import raw data")
    importing_arg_parser.add_argument("--index-file",
                                      dest="index_file",
                                      type=str,
                                      help="index of input media, one file per line")
    importing_arg_parser.add_argument("--annotation-dir",
                                      dest="anno",
                                      type=str,
                                      required=False,
                                      help="corresponding annotation folder")
    importing_arg_parser.add_argument("--gen-dir", dest="gen", type=str, help="storage path of generated data files")
    importing_arg_parser.add_argument("--dataset-name",
                                      dest="dataset_name",
                                      type=str,
                                      required=False,
                                      help="name of the dataset to be created, use tid if not set.")
    importing_arg_parser.add_argument("--src-revs", dest="src_revs", type=str, help="rev: source rev")
    importing_arg_parser.add_argument("--dst-rev",
                                      dest="dst_rev",
                                      type=str,
                                      required=True,
                                      help="rev@tid: destination branch name and task id")
    importing_arg_parser.add_argument('-w', dest='work_dir', type=str, required=False, help='working directory')
    importing_arg_parser.add_argument('--ignore-unknown-types',
                                      dest='ignore_unknown_types',
                                      required=False,
                                      action='store_true',
                                      help='ignore unknown type names in annotation files')
    importing_arg_parser.set_defaults(func=CmdImport)



import argparse
import logging
from typing import Any, List

from google.protobuf import json_format

from mir.commands import base
from mir.protos import mir_command_pb2 as mirpb
from mir.tools import checker, class_ids, context, mir_storage, mir_storage_ops, revs_parser
from mir.tools.code import MirCode


class CmdShow(base.BaseCommand):
    def run(self) -> int:
        logging.debug("command show: %s", self.args)

        return CmdShow.run_with_args(mir_root=self.args.mir_root,
                                     src_revs=self.args.src_revs,
                                     verbose=self.args.verbose)

    @classmethod
    def run_with_args(cls, mir_root: str, src_revs: str, verbose: bool) -> int:
        # check args
        src_typ_rev_tid = revs_parser.parse_single_arg_rev(src_revs, need_tid=False)
        check_code = checker.check(mir_root,
                                   [checker.Prerequisites.IS_INSIDE_MIR_REPO, checker.Prerequisites.HAVE_LABELS])
        if check_code != MirCode.RC_OK:
            return check_code

        # show infos
        cls._show_cis(mir_root, src_typ_rev_tid, verbose)
        cls._show_cks(mir_root, src_typ_rev_tid, verbose)
        cls._show_general(mir_root, src_typ_rev_tid, verbose)

        return MirCode.RC_OK

    @classmethod
    def _show_general(cls, mir_root: str, src_typ_rev_tid: revs_parser.TypRevTid, verbose: bool) -> None:
        cls._show_general_context_config(mir_root=mir_root)

        [metadatas, annotations, _, tasks,
         context] = mir_storage_ops.MirStorageOps.load_multiple_storages(mir_root=mir_root,
                                                                         mir_branch=src_typ_rev_tid.rev,
                                                                         mir_task_id=src_typ_rev_tid.tid,
                                                                         ms_list=mir_storage.get_all_mir_storage(),
                                                                         as_dict=False)
        cls._show_general_metadatas(metadatas)
        cls._show_general_annotations(annotations)
        cls._show_general_context(context)
        cls._show_general_tasks(tasks, verbose)

    @classmethod
    def _show_general_metadatas(cls, mir_metadatas: mirpb.MirMetadatas) -> None:
        un_tr_va_te_counts = [0, 0, 0, 0]
        for _, asset_attr in mir_metadatas.attributes.items():
            un_tr_va_te_counts[asset_attr.tvt_type] += 1

        # if use logging.info here, will cause error output when use mir show in linux pipe
        print(f"metadatas.mir: {len(mir_metadatas.attributes)} assets"
              f" (training: {un_tr_va_te_counts[mirpb.TvtTypeTraining]},"
              f" validation: {un_tr_va_te_counts[mirpb.TvtTypeValidation]},"
              f" test: {un_tr_va_te_counts[mirpb.TvtTypeTest]},"
              f" others: {un_tr_va_te_counts[mirpb.TvtTypeUnknown]})")

    @classmethod
    def _show_general_annotations(cls, mir_annotations: mirpb.MirAnnotations) -> None:
        hid = mir_annotations.head_task_id
        print(f"annotations.mir: hid: {hid}," f" {len(mir_annotations.task_annotations[hid].image_annotations)} assets")

    @classmethod
    def _show_general_context(cls, mir_context: mirpb.MirContext) -> None:
        print(f"context.mir: negative assets cnt: {mir_context.negative_images_cnt}")
        print(f"    project negative assets cnt: {mir_context.project_negative_images_cnt}")
        print(f"    total assets cnt: {mir_context.images_cnt}")

    @classmethod
    def _show_general_context_config(cls, mir_root: str) -> None:
        project_class_ids = context.load(mir_root=mir_root)
        if project_class_ids:
            class_id_and_names: List[str] = []
            class_ids_mgr = class_ids.ClassIdManager(mir_root=mir_root)
            for class_id in project_class_ids:
                class_id_and_names.append(f"{class_id} ({class_ids_mgr.main_name_for_id(class_id)})")
            print(f"project classes: {'; '.join(class_id_and_names)}")
        else:
            print('project classes: none')

    @classmethod
    def _show_general_tasks(cls, mir_tasks: mirpb.MirTasks, verbose: bool) -> None:
        hid = mir_tasks.head_task_id
        task = mir_tasks.tasks[hid]
        if not verbose:
            print(f"tasks.mir: hid: {hid}, code: {task.return_code}, error msg: {task.return_msg}\n"
                  f"    model hash: {task.model.model_hash}\n"
                  f"    map: {task.model.mean_average_precision}\n"
                  f"    executor: {task.executor}")
        else:
            print(f"tasks.mir: {json_format.MessageToDict(mir_tasks, preserving_proto_field_name=True)}")

    @classmethod
    def _show_cis(cls, mir_root: str, src_typ_rev_tid: revs_parser.TypRevTid, verbose: bool) -> None:
        mir_context: mirpb.MirContext = mir_storage_ops.MirStorageOps.load_single_storage(
            mir_root=mir_root, mir_branch=src_typ_rev_tid.rev, mir_task_id=src_typ_rev_tid.tid, ms=mirpb.MIR_CONTEXT)
        cls_id_mgr = class_ids.ClassIdManager(mir_root=mir_root)
        if verbose:
            print('predefined key ids and assets count:')
            cls._show_cis_verbose(predefined_keyids_cnt=mir_context.predefined_keyids_cnt, cls_id_mgr=cls_id_mgr)
            if mir_context.project_predefined_keyids_cnt:
                print('project predefined key ids and assets count:')
                cls._show_cis_verbose(predefined_keyids_cnt=mir_context.project_predefined_keyids_cnt,
                                      cls_id_mgr=cls_id_mgr)
        else:
            type_names = [cls_id_mgr.main_name_for_id(ci) or '' for ci in mir_context.predefined_keyids_cnt.keys()]
            print(';'.join(type_names))

    @classmethod
    def _show_cis_verbose(cls, predefined_keyids_cnt: Any, cls_id_mgr: class_ids.ClassIdManager) -> None:
        for ci, cnt in predefined_keyids_cnt.items():
            main_name = cls_id_mgr.main_name_for_id(ci)
            if main_name:
                print(f"  {main_name}: {cnt}")
            else:
                print(f"  {ci} (unknown ci): {cnt}")

    @classmethod
    def _show_cks(cls, mir_root: str, src_typ_rev_tid: revs_parser.TypRevTid, verbose: bool) -> None:
        print('')  # currently no customized keywords


def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: argparse.ArgumentParser) -> None:
    show_arg_parser = subparsers.add_parser('show',
                                            parents=[parent_parser],
                                            description='use this command to show current workspace informations',
                                            help='show current workspace informations')
    show_arg_parser.add_argument('--verbose', dest='verbose', action='store_true', help='show verbose info')
    show_arg_parser.add_argument('--src-revs', dest='src_revs', type=str, help='rev@bid: source rev and base task id')
    show_arg_parser.set_defaults(func=CmdShow)



PRODUCER_KEY = 'producer'
PRODUCER_NAME = 'ymir'
EXECUTOR_CONFIG_KEY = 'executor_config'
TASK_CONTEXT_KEY = 'task_context'
TASK_CONTEXT_PARAMETERS_KEY = 'task_parameters'
EXECUTOR_OUTLOG_NAME = 'ymir-executor-out.log'



from collections import defaultdict
import logging
import os
from typing import Dict, List, Set, Tuple

import xml.dom.minidom

from mir.tools import class_ids
from mir.tools.code import MirCode
from mir.tools.errors import MirRuntimeError
from mir.tools.phase_logger import PhaseLoggerCenter
from mir.protos import mir_command_pb2 as mirpb


def _get_dom_xml_tag_node(node: xml.dom.minidom.Element, tag_name: str) -> xml.dom.minidom.Element:
    """
    suppose we have the following xml:
    ```
    <blabla>
        <tag1>tag1_value</tag1>
        <tag2>tag2_value</tag2>
    </blabla>
    ```
    and we have node point to <blabla>, we can use this function to get node tag1 and tag2 \n
    if tag not found, returns None
    """
    tag_nodes = node.getElementsByTagName(tag_name)
    if len(tag_nodes) > 0 and len(tag_nodes[0].childNodes) > 0:
        return tag_nodes[0]
    raise MirRuntimeError(MirCode.RC_CMD_INVALID_FILE, f"found no element for key: {tag_name}")


def _get_dom_xml_tag_data(node: xml.dom.minidom.Element, tag_name: str) -> str:
    """
    suppose we have the following xml:
    ```
    <blabla>
        <tag1>tag1_value</tag1>
        <tag2>tag2_value</tag2>
    </blabla>
    ```
    and we have node point to <blabla>, we can use this function to get tag1_value and tag2_value \n
    if tag not found, returns empty str
    """
    tag_node = _get_dom_xml_tag_node(node, tag_name)
    if tag_node and len(tag_node.childNodes) > 0:
        return tag_node.childNodes[0].data
    return ''


def _xml_obj_to_annotation(obj: xml.dom.minidom.Element,
                           class_type_manager: class_ids.ClassIdManager) -> mirpb.Annotation:
    """
    generate mirpb.Annotation instance from object node in coco and pascal annotation xml file
    """
    name = _xml_obj_to_type_name(obj)
    bndbox_node = _get_dom_xml_tag_node(obj, "bndbox")
    xmin = int(float(_get_dom_xml_tag_data(bndbox_node, "xmin")))
    ymin = int(float(_get_dom_xml_tag_data(bndbox_node, "ymin")))
    xmax = int(float(_get_dom_xml_tag_data(bndbox_node, "xmax")))
    ymax = int(float(_get_dom_xml_tag_data(bndbox_node, "ymax")))
    width = xmax - xmin + 1
    height = ymax - ymin + 1

    annotation = mirpb.Annotation()
    annotation.class_id = class_type_manager.id_and_main_name_for_name(name)[0]
    annotation.box.x = xmin
    annotation.box.y = ymin
    annotation.box.w = width
    annotation.box.h = height
    annotation.score = 0
    return annotation


def _xml_obj_to_type_name(obj: xml.dom.minidom.Element) -> str:
    return _get_dom_xml_tag_data(obj, "name").lower()


def _read_customized_Keywords(ck_file: str) -> Dict[str, Set[str]]:
    """
    read customized keywords out from ck file

    Args:
        ck_file (str): tsv file, <asset_name>\t<ck1>\t<ck2>...

    Returns:
        Dict[str, Set[str]]: key: asset name, value: set of customized keywords

    Raises:
        ValueError: if found dumplicat key in ck file
    """
    name_cks = {}  # type: Dict[str, Set[str]]
    if not ck_file:
        return name_cks

    with open(ck_file, 'r') as f:
        lines = f.read().splitlines()

    for line in lines:
        if not line:
            continue
        components = line.split('\t')
        if len(components) == 1:
            continue

        if components[0] in name_cks:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                                  error_message=f'dumplicate asset name in ck file: {components[0]}')

        name_cks[components[0]] = set(components[1:])

    return name_cks


def import_annotations(mir_annotation: mirpb.MirAnnotations, mir_keywords: mirpb.MirKeywords, in_sha1_file: str,
                       ck_file: str, mir_root: str, annotations_dir_path: str, task_id: str,
                       phase: str) -> Tuple[int, Dict[str, int]]:
    """
    imports annotations

    Args:
        mir_annotation (mirpb.MirAnnotations): data buf for annotations.mir
        mir_keywords (mirpb.MirKeywords): data buf for keywords.mir
        in_sha1_file (str): path to sha1 file
        ck_file (str): path to customized keywords file
        mir_root (str): path to mir repo
        annotations_dir_path (str): path to annotations root
        task_id (str): task id
        phase (str): process phase

    Returns:
        Tuple[int, Dict[str, int]]: return code and unknown type names
    """
    unknown_types_and_count: Dict[str, int] = defaultdict(int)

    if not in_sha1_file:
        logging.error("empty sha1_file")
        return MirCode.RC_CMD_INVALID_ARGS, unknown_types_and_count

    # read type_id_name_dict and type_name_id_dict
    class_type_manager = class_ids.ClassIdManager(mir_root=mir_root)
    logging.info("loaded type id and names: %d", class_type_manager.size())

    # read customized keywords from ck_file
    cks = _read_customized_Keywords(ck_file)

    image_annotations = mir_annotation.task_annotations[task_id].image_annotations

    assethash_filename_list = []  # type: List[Tuple[str, str, str]]
    with open(in_sha1_file, "r") as in_file:
        for line in in_file.readlines():
            line_components = line.strip().split('\t')
            if not line_components or len(line_components) < 2:
                logging.warning("incomplete line: %s", line)
                continue
            asset_hash, file_name = line_components[0], line_components[1]
            main_file_name = os.path.splitext(os.path.basename(file_name))[0]
            assethash_filename_list.append((asset_hash, main_file_name, file_name))

    total_assethash_count = len(assethash_filename_list)
    logging.info(f"wrting {total_assethash_count} annotations")

    counter = 0
    missing_annotations_counter = 0
    for asset_hash, main_file_name, file_path in assethash_filename_list:
        # for each asset, import it's annotations
        annotation_file = os.path.join(annotations_dir_path, main_file_name + '.xml') if annotations_dir_path else None
        if not annotation_file or not os.path.isfile(annotation_file):
            missing_annotations_counter += 1
        else:
            # if have annotation file, import annotations and predefined key ids
            dom_tree = xml.dom.minidom.parse(annotation_file)
            if not dom_tree:
                logging.error(f"cannot open annotation_file: {annotation_file}")
                return MirCode.RC_CMD_INVALID_ARGS, unknown_types_and_count

            collection = dom_tree.documentElement
            objects = collection.getElementsByTagName("object")
            for idx, obj in enumerate(objects):
                type_name = _xml_obj_to_type_name(obj)
                if class_type_manager.has_name(type_name):
                    annotation = _xml_obj_to_annotation(obj, class_type_manager)
                    annotation.index = idx
                    image_annotations[asset_hash].annotations.append(annotation)
                else:
                    unknown_types_and_count[type_name] += 1

        # import customized keywords
        if file_path in cks:
            mir_keywords.keywords[asset_hash].customized_keywords[:] = cks[file_path]

        counter += 1
        if counter % 5000 == 0:
            PhaseLoggerCenter.update_phase(phase=phase, local_percent=(counter / total_assethash_count))

    if missing_annotations_counter > 0:
        logging.warning(f"asset count that have no annotations: {missing_annotations_counter}")

    return MirCode.RC_OK, unknown_types_and_count



"""
use this module to read contents in other branch head ref or from other tags \n
some mir commands, such as `mir search`, `mir merge` will use this module
"""

import io

from mir import scm
from mir.tools.code import MirCode
from mir.tools.errors import MirRuntimeError


def read_mir(mir_root: str, rev: str, file_name: str) -> bytes:
    if not mir_root or not file_name or not rev:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='invalid args')

    scm_git = scm.Scm(mir_root if mir_root else ".", scm_executable="git")
    blob_hash = scm_git.rev_parse(f"{rev}:{file_name}")
    if not blob_hash:
        raise MirRuntimeError(MirCode.RC_CMD_INVALID_MIR_REPO, f"found no file: {rev}:{file_name}")

    bio = io.BytesIO()
    scm_git.cat_file(['-p', blob_hash], output_stream=bio)
    return bio.getvalue()



from enum import IntEnum


class MirCode(IntEnum):
    RC_OK = 0  # everything is ok, command finished without any errors or warnings
    RC_CMD_INVALID_ARGS = 160001  # lack of necessary args, or unexpected args
    RC_CMD_EMPTY_TRAIN_SET = 160002  # no training set when training
    RC_CMD_EMPTY_VAL_SET = 160003  # no validation set when training
    RC_CMD_CONTAINER_ERROR = 160004  # container errors
    RC_CMD_UNKNOWN_TYPES = 160005  # unknown types found, and can not be ignored when mir import
    RC_CMD_INVALID_BRANCH_OR_TAG = 160006  # invalid branch name or tag name
    RC_CMD_DIRTY_REPO = 160007  # repo is dirty when mir commit
    RC_CMD_MERGE_ERROR = 160008  # error occured when mir merge
    RC_CMD_INVALID_MIR_REPO = 160009
    RC_CMD_INVALID_FILE = 160010
    RC_CMD_NO_RESULT = 160011  # no result for training, mining and infer
    RC_CMD_ERROR_UNKNOWN = 169999



"""mir command line interface"""
import argparse
import logging
import os
import sys
from typing import Any, cast, Protocol

from mir import version
from mir.commands import (init, branch, checkout, commit, copy, exporting, filter, log, merge, reset, sampling, show,
                          status, training, mining, importing, infer, model_importing)

_COMMANDS_ = [
    init, branch, checkout, commit, copy, exporting, filter, log, merge, reset, sampling, show, status, training,
    mining, importing, infer, model_importing
]


class _CmdModuleProtocol(Protocol):
    """
    protocol for all modules in `_COMMANDS_`, to cheat mypy
    """
    @staticmethod
    def bind_to_subparsers(sub_parsers: argparse.Action, share_parser: argparse.ArgumentParser) -> None:
        pass


class MirParser(argparse.ArgumentParser):
    """Custom parser class for mir CLI."""
    def error(self, message: str, cmd_cls: str = None) -> Any:  # pylint: disable=arguments-differ
        logging.error(message)

    def parse_args(self, args: Any = None, namespace: Any = None) -> Any:
        # NOTE: overriding to provide a more granular help message.
        args, argv = self.parse_known_args(args, namespace)
        if argv:
            msg = "unrecognized arguments: %s"
            self.error(msg % " ".join(argv), getattr(args, "func", None))
        return args


class VersionAction(argparse.Action):
    """Show mir version and exits"""
    def __call__(self, parser: Any, namespace: Any, values: Any, option_string: Any = None) -> None:
        logging.info("mir version: {0}".format(version.__version__))
        sys.exit(0)


class DebugModeAction(argparse.Action):
    """Work in debug mode, and print debug info."""
    def __call__(self, parser: Any, namespace: Any, values: Any, option_string: Any = None) -> None:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

        logging.basicConfig(stream=sys.stdout,
                            format='%(levelname)-8s: [%(asctime)s] %(filename)s:%(lineno)s:%(funcName)s(): %(message)s',
                            datefmt='%Y%m%d-%H:%M:%S',
                            level=logging.DEBUG)
        logging.debug("in debug mode")


def create_main_parser() -> argparse.ArgumentParser:
    parent_parser = argparse.ArgumentParser()
    main_parser = MirParser(
        prog="mir",
        description="MIR",
        parents=[parent_parser],
        formatter_class=argparse.RawTextHelpFormatter,
        add_help=False,
    )

    main_parser.add_argument("-v", "--version", nargs=0, action=VersionAction, help="show program's version")

    main_parser.add_argument("-d", "--debug", nargs=0, action=DebugModeAction, help="set debug mode, print more infos.")

    share_parser = argparse.ArgumentParser(add_help=False)
    share_parser.add_argument("--root",
                              "-r",
                              dest="mir_root",
                              type=str,
                              default=os.getcwd(),
                              help="root path to the mir repo, use . if not set.")

    # sub parsers
    subparsers = main_parser.add_subparsers(title="sub commands",
                                            metavar="COMMAND",
                                            dest="cmd",
                                            help="Use mir COMMAND --help for sub command help")

    for command in _COMMANDS_:
        cast(_CmdModuleProtocol, command).bind_to_subparsers(subparsers, share_parser)

    return main_parser



from typing import Any, List

from mir.protos import mir_command_pb2 as mirpb


MIR_ASSOCIATED_FILES = ['.git', '.mir', '.gitignore', '.mir_lock']


def mir_type(mir_storage: 'mirpb.MirStorage.V') -> Any:
    MIR_TYPE = {
        mirpb.MirStorage.MIR_METADATAS: mirpb.MirMetadatas,
        mirpb.MirStorage.MIR_ANNOTATIONS: mirpb.MirAnnotations,
        mirpb.MirStorage.MIR_KEYWORDS: mirpb.MirKeywords,
        mirpb.MirStorage.MIR_TASKS: mirpb.MirTasks,
        mirpb.MirStorage.MIR_CONTEXT: mirpb.MirContext,
    }
    return MIR_TYPE[mir_storage]


def mir_path(mir_storage: 'mirpb.MirStorage.V') -> str:
    MIR_PATH = {
        mirpb.MirStorage.MIR_METADATAS: 'metadatas.mir',
        mirpb.MirStorage.MIR_ANNOTATIONS: 'annotations.mir',
        mirpb.MirStorage.MIR_KEYWORDS: 'keywords.mir',
        mirpb.MirStorage.MIR_TASKS: 'tasks.mir',
        mirpb.MirStorage.MIR_CONTEXT: 'context.mir',
    }
    return MIR_PATH[mir_storage]


def mir_attr_name(mir_storage: 'mirpb.MirStorage.V') -> str:
    MIR_STORAGE_TO_ATTR_NAME = {
        mirpb.MirStorage.MIR_METADATAS: 'mir_metadatas',
        mirpb.MirStorage.MIR_ANNOTATIONS: 'mir_annotations',
        mirpb.MirStorage.MIR_KEYWORDS: 'mir_keywords',
        mirpb.MirStorage.MIR_TASKS: 'mir_tasks',
        mirpb.MirStorage.MIR_CONTEXT: 'mir_context',
    }
    return MIR_STORAGE_TO_ATTR_NAME[mir_storage]


def get_all_mir_paths() -> List[str]:
    return [mir_path(ms) for ms in get_all_mir_storage()]


def get_all_mir_storage() -> List['mirpb.MirStorage.V']:
    return [
        mirpb.MirStorage.MIR_METADATAS,
        mirpb.MirStorage.MIR_ANNOTATIONS,
        mirpb.MirStorage.MIR_KEYWORDS,
        mirpb.MirStorage.MIR_TASKS,
        mirpb.MirStorage.MIR_CONTEXT,
    ]



"""
exports the assets and annotations from mir format to ark-training-format
"""

from collections.abc import Collection
from enum import Enum
import logging
import json
import os
from typing import Any, Callable, Dict, List, Optional, Set
import uuid
import xml.etree.ElementTree as ElementTree

from mir.protos import mir_command_pb2 as mirpb
from mir.tools import class_ids, mir_storage_ops
from mir.tools import utils as mir_utils
from mir.tools.code import MirCode
from mir.tools.errors import MirRuntimeError


class ExportError(Exception):
    """
    exception type raised by function `export`
    """
    pass


class ExportFormat(str, Enum):
    EXPORT_FORMAT_UNKNOWN = 'unknown'
    EXPORT_FORMAT_NO_ANNOTATION = 'none'
    EXPORT_FORMAT_ARK = 'ark'
    EXPORT_FORMAT_VOC = 'voc'
    EXPORT_FORMAT_LS_JSON = 'ls_json'  # label studio json format


def check_support_format(anno_format: str) -> bool:
    return anno_format in support_format_type()


def support_format_type() -> List[str]:
    return [f.value for f in ExportFormat]


def format_type_from_str(anno_format: str) -> ExportFormat:
    return ExportFormat(anno_format.lower())


def format_file_ext(anno_format: ExportFormat) -> str:
    _format_ext_map = {
        ExportFormat.EXPORT_FORMAT_ARK: '.txt',
        ExportFormat.EXPORT_FORMAT_VOC: '.xml',
        ExportFormat.EXPORT_FORMAT_LS_JSON: '.json',
    }
    return _format_ext_map[anno_format]


def _rel_annotation_path_for_asset(rel_asset_path: str, format_type: ExportFormat) -> str:
    rel_asset_path_without_ext = os.path.splitext(rel_asset_path)[0]
    return f"{rel_asset_path_without_ext}{format_file_ext(format_type)}"


def format_file_output_func(anno_format: ExportFormat) -> Callable:
    _format_func_map = {
        ExportFormat.EXPORT_FORMAT_ARK: _single_image_annotations_to_ark,
        ExportFormat.EXPORT_FORMAT_VOC: _single_image_annotations_to_voc,
        ExportFormat.EXPORT_FORMAT_LS_JSON: _single_image_annotations_to_ls_json,
    }
    return _format_func_map[anno_format]


# public: export
def export(mir_root: str,
           assets_location: str,
           class_type_ids: Dict[int, int],
           asset_ids: Set[str],
           asset_dir: str,
           annotation_dir: str,
           need_ext: bool,
           need_id_sub_folder: bool,
           base_branch: str,
           base_task_id: str,
           format_type: ExportFormat,
           index_file_path: str = '',
           index_assets_prefix: str = '',
           index_annotations_prefix: str = '') -> bool:
    """
    export assets and annotations

    Args:
        mir_root (str): path to mir repo root directory
        assets_location (str): path to assets storage directory
        class_type_ids (Dict[int, int]): class ids (and it's mapping value)
            all objects within this dict keys will be exported, if None, export everything;
        asset_ids (Set[str]): export asset ids
        asset_dir (str): asset directory
        annotation_dir (str): annotation directory, if format_type is NO_ANNOTATION, this could be None
        need_ext (bool): if true, all export assets will have it's type as ext, jpg, png, etc.
        need_id_sub_folder (bool): if True, use last 2 chars of asset id as a sub folder name
        base_branch (str): data branch
        format_type (ExportFormat): format type, NONE means exports no annotations
        index_file_path (str): path to index file, if None, generates no index file
        index_assets_prefix (str): prefix path added to each asset index path
        index_annotations_prefix (str): prefix path added to each annotation index path

    Raises:
        MirRuntimeError

    Returns:
        bool: returns True if success
    """
    if not mir_root:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                              error_message="invalid mir_repo")

    if not check_support_format(format_type):
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                              error_message=f"invalid --format: {format_type}")

    # export assets
    os.makedirs(asset_dir, exist_ok=True)
    asset_result = mir_utils.store_assets_to_dir(asset_ids=list(asset_ids),
                                                 out_root=asset_dir,
                                                 sub_folder=".",
                                                 asset_location=assets_location,
                                                 overwrite=False,
                                                 create_prefix=need_id_sub_folder,
                                                 need_suffix=need_ext)

    # export annotations
    if format_type != ExportFormat.EXPORT_FORMAT_NO_ANNOTATION:
        [mir_metadatas, mir_annotations] = mir_storage_ops.MirStorageOps.load_multiple_storages(
            mir_root=mir_root,
            mir_branch=base_branch,
            mir_task_id=base_task_id,
            ms_list=[mirpb.MirStorage.MIR_METADATAS, mirpb.MirStorage.MIR_ANNOTATIONS])

        # add all annotations to assets_to_det_annotations_dict
        # key: asset_id as str, value: annotations as List[mirpb.Annotation]
        assets_to_det_annotations_dict = _annotations_by_assets(
            mir_annotations=mir_annotations,
            class_type_ids=set(class_type_ids.keys()) if class_type_ids else None,
            base_task_id=mir_annotations.head_task_id)

        _export_detect_annotations_to_path(asset_ids=list(asset_ids),
                                           format_type=format_type,
                                           mir_metadatas=mir_metadatas,
                                           annotations_dict=assets_to_det_annotations_dict,
                                           class_type_mapping=class_type_ids,
                                           dest_path=annotation_dir,
                                           mir_root=mir_root,
                                           assert_id_filename_map=asset_result)

    # generate index file
    if index_file_path:
        _generate_asset_index_file(asset_rel_paths=asset_result.values(),
                                   index_assets_prefix=index_assets_prefix,
                                   index_annotations_prefix=index_annotations_prefix,
                                   index_file_path=index_file_path,
                                   format_type=format_type)

    return True


def _generate_asset_index_file(asset_rel_paths: Collection,
                               index_assets_prefix: str,
                               index_annotations_prefix: str,
                               index_file_path: str,
                               format_type: ExportFormat,
                               overwrite: bool = True,
                               image_exts: tuple = ('.jpg', '.jpeg', '.png')) -> None:
    """
    generate index file for export result

    if format_type == NO_ANNOTATION, index file contains only asset paths

    if not, index file contains both asset and annotation paths, separated by `\t`

    Args:
        asset_rel_paths (Collection): the relative asset paths, element type: str
        index_assets_prefix (str): prefix path added in front of each element in asset_rel_paths
        index_annotations_prefix (str): prefix path added in front of each annotations
        index_file_path (str): index file save path
        format_type (ExporterFormat): format type
        override (bool): if True, override if file already exists, if False, raise Exception when already exists

    Raise:
        MirRuntimeError: if index file already exists, and override set to False
    """
    if not asset_rel_paths or not index_file_path:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                              error_message='empty asset_rel_paths or index_file_path')
    if os.path.exists(index_file_path):
        if overwrite:
            logging.warning(f"index file already exists, overwriting: {index_file_path}")
        else:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                                  error_message=f"index file already exists: {index_file_path}")

    with open(index_file_path, 'w') as f:
        for item in asset_rel_paths:
            if os.path.splitext(item)[1] not in image_exts:
                logging.warning(f"unsupported image ext in path: {item}")
                continue

            asset_path = os.path.join(index_assets_prefix, item)
            if format_type == ExportFormat.EXPORT_FORMAT_NO_ANNOTATION:
                annotation_path = ''
            else:
                annotation_rel_path = _rel_annotation_path_for_asset(rel_asset_path=item, format_type=format_type)
                annotation_path = os.path.join(index_annotations_prefix, annotation_rel_path)
            f.write(f"{asset_path}\t{annotation_path}\n")


# private: export annotations: general
def _annotations_by_assets(mir_annotations: mirpb.MirAnnotations, class_type_ids: Optional[Set[int]],
                           base_task_id: str) -> Dict[str, List[mirpb.Annotation]]:
    """
    get annotations by assets

    Args:
        mir_annotations (mirpb.MirAnnotations): annotations
        class_type_ids (Optional[Set[int]]): only type ids within it could be output, if None, no id filter applied
        base_task_id (str): base task id

    Returns:
        Dict, key: asset id, value: List[mirpb.Annotation]
    """
    assets_to_det_annotations_dict = {}  # type: Dict[str, List[mirpb.Annotation]]

    if base_task_id not in mir_annotations.task_annotations:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_MIR_REPO,
                              error_message=f"base task id: {base_task_id} not in mir_annotations")

    task_annotations = mir_annotations.task_annotations[base_task_id]
    for asset_id, image_annotations in task_annotations.image_annotations.items():
        matched_annotations = [
            annotation for annotation in image_annotations.annotations
            if (not class_type_ids or (annotation.class_id in class_type_ids))
        ]
        assets_to_det_annotations_dict[asset_id] = matched_annotations

    return assets_to_det_annotations_dict


def _export_detect_annotations_to_path(asset_ids: List[str], format_type: ExportFormat,
                                       mir_metadatas: mirpb.MirMetadatas,
                                       annotations_dict: Dict[str, List[mirpb.Annotation]],
                                       class_type_mapping: Optional[Dict[int, int]], dest_path: str, mir_root: str,
                                       assert_id_filename_map: Dict[str, str]) -> None:
    if not asset_ids:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='empty asset_ids')
    if not mir_metadatas:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='invalid mir_metadatas')

    os.makedirs(dest_path, exist_ok=True)

    cls_id_mgr = class_ids.ClassIdManager(mir_root=mir_root)

    missing_counter = 0
    empty_counter = 0
    for asset_id in asset_ids:
        if asset_id not in mir_metadatas.attributes:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                                  error_message=f"can not find asset id: {asset_id} in mir_metadatas")
        attrs = mir_metadatas.attributes[asset_id]

        if asset_id not in annotations_dict:
            missing_counter += 1
            annotations = []
        else:
            annotations = annotations_dict[asset_id]
        if len(annotations) == 0:
            empty_counter += 1

        format_func = format_file_output_func(anno_format=format_type)
        asset_file_name = assert_id_filename_map[asset_id]
        anno_str = format_func(asset_id=asset_id,
                               attrs=attrs,
                               annotations=annotations,
                               class_type_mapping=class_type_mapping,
                               cls_id_mgr=cls_id_mgr,
                               asset_filename=asset_file_name)

        annotation_file_path = os.path.join(
            dest_path, _rel_annotation_path_for_asset(rel_asset_path=asset_file_name, format_type=format_type))
        os.makedirs(os.path.dirname(annotation_file_path), exist_ok=True)
        with open(annotation_file_path, 'w') as f:
            f.write(anno_str)

    logging.info(f"missing annotations: {missing_counter}, "
                 f"empty annotations: {empty_counter} out of {len(asset_ids)} assets")


def _single_image_annotations_to_ark(asset_id: str, attrs: Any, annotations: List[mirpb.Annotation],
                                     class_type_mapping: Optional[Dict[int, int]], cls_id_mgr: class_ids.ClassIdManager,
                                     asset_filename: str) -> str:
    output_str = ""
    for annotation in annotations:
        mapped_id = class_type_mapping[annotation.class_id] if class_type_mapping else annotation.class_id
        output_str += f"{mapped_id}, {annotation.box.x}, {annotation.box.y}, "
        output_str += f"{annotation.box.x + annotation.box.w - 1}, {annotation.box.y + annotation.box.h - 1}\n"
    return output_str


def _single_image_annotations_to_voc(asset_id: str, attrs: Any, annotations: List[mirpb.Annotation],
                                     class_type_mapping: Optional[Dict[int, int]], cls_id_mgr: class_ids.ClassIdManager,
                                     asset_filename: str) -> str:
    # annotation
    annotation_node = ElementTree.Element('annotation')

    # annotation: folder
    folder_node = ElementTree.SubElement(annotation_node, 'folder')
    folder_node.text = 'folder'

    # annotation: filename
    filename_node = ElementTree.SubElement(annotation_node, 'filename')
    filename_node.text = asset_filename

    # annotation: source
    source_node = ElementTree.SubElement(annotation_node, 'source')

    # annotation: source: database
    database_node = ElementTree.SubElement(source_node, 'database')
    database_node.text = attrs.dataset_name or 'unknown'

    # annotation: source: annotation
    annotation2_node = ElementTree.SubElement(source_node, 'annotation')
    annotation2_node.text = 'unknown'

    # annotation: source: image
    image_node = ElementTree.SubElement(source_node, 'image')
    image_node.text = 'unknown'

    # annotation: size
    size_node = ElementTree.SubElement(annotation_node, 'size')

    # annotation: size: width
    width_node = ElementTree.SubElement(size_node, 'width')
    width_node.text = str(attrs.width)

    # annotation: size: height
    height_node = ElementTree.SubElement(size_node, 'height')
    height_node.text = str(attrs.height)

    # annotation: size: depth
    depth_node = ElementTree.SubElement(size_node, 'depth')
    depth_node.text = str(attrs.image_channels)

    # annotation: segmented
    segmented_node = ElementTree.SubElement(annotation_node, 'segmented')
    segmented_node.text = '0'

    # annotation: object(s)
    for annotation in annotations:
        object_node = ElementTree.SubElement(annotation_node, 'object')

        name_node = ElementTree.SubElement(object_node, 'name')
        name_node.text = cls_id_mgr.main_name_for_id(annotation.class_id) or 'unknown'

        pose_node = ElementTree.SubElement(object_node, 'pose')
        pose_node.text = 'unknown'

        truncated_node = ElementTree.SubElement(object_node, 'truncated')
        truncated_node.text = 'unknown'

        occluded_node = ElementTree.SubElement(object_node, 'occluded')
        occluded_node.text = '0'

        bndbox_node = ElementTree.SubElement(object_node, 'bndbox')

        xmin_node = ElementTree.SubElement(bndbox_node, 'xmin')
        xmin_node.text = str(annotation.box.x)

        ymin_node = ElementTree.SubElement(bndbox_node, 'ymin')
        ymin_node.text = str(annotation.box.y)

        xmax_node = ElementTree.SubElement(bndbox_node, 'xmax')
        xmax_node.text = str(annotation.box.x + annotation.box.w - 1)

        ymax_node = ElementTree.SubElement(bndbox_node, 'ymax')
        ymax_node.text = str(annotation.box.y + annotation.box.h - 1)

        difficult_node = ElementTree.SubElement(object_node, 'difficult')
        difficult_node.text = '0'

    return ElementTree.tostring(element=annotation_node, encoding='unicode')


def _single_image_annotations_to_ls_json(asset_id: str, attrs: Any, annotations: List[mirpb.Annotation],
                                         class_type_mapping: Optional[Dict[int, int]],
                                         cls_id_mgr: class_ids.ClassIdManager, asset_filename: str) -> str:
    out_type = "predictions"  # out_type: annotation type - "annotations" or "predictions"
    to_name = 'image'  # to_name: object name from Label Studio labeling config
    from_name = 'label'  # control tag name from Label Studio labeling config
    task: Dict = {
        out_type: [{
            "result": [],
            "ground_truth": False,
        }],
        "data": {
            "image": asset_filename
        }
    }

    for annotation in annotations:
        bbox_x, bbox_y = float(annotation.box.x), float(annotation.box.y)
        bbox_width, bbox_height = float(annotation.box.w), float(annotation.box.h)
        img_width, img_height = attrs.width, attrs.height
        item = {
            "id": uuid.uuid4().hex[0:10],  # random id to identify this annotation.
            "type": "rectanglelabels",
            "value": {
                # Units of image annotations in label studio is percentage of image width/height.
                # https://labelstud.io/guide/predictions.html#Units-of-image-annotations
                "x": bbox_x / img_width * 100,
                "y": bbox_y / img_height * 100,
                "width": bbox_width / img_width * 100,
                "height": bbox_height / img_height * 100,
                "rotation": 0,
                "rectanglelabels": [cls_id_mgr.main_name_for_id(annotation.class_id) or 'unknown']
            },
            "to_name": to_name,
            "from_name": from_name,
            "image_rotation": 0,
            "original_width": img_width,
            "original_height": img_height
        }
        task[out_type][0]['result'].append(item)
    return json.dumps(task)



import argparse
import json
import logging
import os
import subprocess
import time
from typing import Any, Tuple, Optional

import yaml

from mir.commands import base
from mir.tools import settings as mir_settings, utils as mir_utils
from mir.tools.code import MirCode
from mir.tools.errors import MirRuntimeError


class CmdInfer(base.BaseCommand):
    """
    infer command

    Steps:
        a. prepare_env: make dirs
        b. prepare_assets: copy assets in orig index.tsv into work_dir/in/candidate, and make candidate index.tsv
        c. prepare_model: copy model to work_dir/in/models and unpack
        d. prepare_config_file: generate work_dir/in/config.yaml
        e. run_docker_cmd: bind paths and run docker cmd

    About path bindings:
        a. work_dir/in/assets or cache -> /in/assets
        b. work_dir/in/models -> /in/models
        c. work_dir/in/candidate-index.tsv -> /in/candidate-index.tsv
        d. work_dir/in/config.yaml -> /in/config.yaml
        e. work_dir/out -> out
    """
    def run(self) -> int:
        logging.debug("command infer: %s", self.args)

        return CmdInfer.run_with_args(work_dir=self.args.work_dir,
                                      media_path=self.args.work_dir,
                                      model_location=self.args.model_location,
                                      model_hash=self.args.model_hash,
                                      index_file=self.args.index_file,
                                      config_file=self.args.config_file,
                                      executor=self.args.executor,
                                      executant_name=self.args.executant_name,
                                      run_infer=True,
                                      run_mining=False)

    @staticmethod
    def run_with_args(work_dir: str,
                      media_path: str,
                      model_location: str,
                      model_hash: str,
                      index_file: str,
                      config_file: str,
                      executor: str,
                      executant_name: str,
                      task_id: str = f"default-infer-{time.time()}",
                      shm_size: str = None,
                      run_infer: bool = False,
                      run_mining: bool = False) -> int:
        """run infer command

        This function can be called from cmd infer, or as part of minig cmd

        Args:
            work_dir (str): work directory, in which the command generates tmp files
            media_path (str): media path, all medias in `index_file` should all in this `media_path`
                in cmd infer, set it to work_dir, in cmd mining, set it to media_cache or work_dir
            model_location (str): model location
            model_hash (str): model package hash (or model package name)
            index_file (str): index file, each line means an image abs path
            config_file (str): configuration file passed to infer executor
            executor (str): docker image name used to infer
            executant_name (str): docker container name
            task_id (str, optional): id of this infer (or mining) task. Defaults to 'default-infer' + timestamp.
            shm_size (str, optional): shared memory size used to start the infer docker. Defaults to None.
            run_infer (bool, optional): run or not run infer. Defaults to False.
            run_mining (bool, optional): run or not run mining, Defaults to False.

        Returns:
            int: [description]
        """
        # check args
        if not work_dir:
            logging.error('empty --work-dir, abort')
            return MirCode.RC_CMD_INVALID_ARGS
        if not model_location:
            logging.error('empty --model-location, abort')
            return MirCode.RC_CMD_INVALID_ARGS
        if not model_hash:
            logging.error('empty --model-hash, abort')
            return MirCode.RC_CMD_INVALID_ARGS
        if not index_file or not os.path.isfile(index_file):
            logging.error(f"invalid --index-file: {index_file}, abort")
            return MirCode.RC_CMD_INVALID_ARGS

        if not config_file:
            logging.error("empty --task-config-file")
            return MirCode.RC_CMD_INVALID_ARGS
        if not os.path.isfile(config_file):
            logging.error(f"invalid --task-config-file {config_file}, not a file, abort")
            return MirCode.RC_CMD_INVALID_ARGS

        if not run_infer and not run_mining:
            logging.warning('invalid run_infer and run_mining: both false')
            return MirCode.RC_OK

        if not executor:
            logging.error('empty --executor, abort')
            return MirCode.RC_CMD_INVALID_ARGS

        if not executant_name:
            executant_name = task_id

        _, work_model_path, work_out_path = _prepare_env(work_dir)
        work_index_file = os.path.join(work_dir, 'in', 'candidate-index.tsv')
        work_config_file = os.path.join(work_dir, 'in', 'config.yaml')
        work_env_config_file = os.path.join(work_dir, 'in', 'env.yaml')

        _prepare_assets(index_file=index_file, work_index_file=work_index_file, media_path=media_path)

        model_storage = mir_utils.prepare_model(model_location=model_location,
                                                model_hash=model_hash,
                                                dst_model_path=work_model_path)
        model_names = model_storage.models
        class_names = model_storage.class_names
        if not class_names:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_FILE,
                                  error_message=f"empty class names in model: {model_hash}")

        with open(config_file, 'r') as f:
            config = yaml.safe_load(f)

        prepare_config_file(config=config,
                            dst_config_file=work_config_file,
                            class_names=class_names,
                            task_id=task_id,
                            model_params_path=[os.path.join('/in/models', name) for name in model_names],
                            run_infer=run_infer,
                            run_mining=run_mining)

        mir_utils.generate_mining_infer_env_config_file(task_id=task_id,
                                                        run_mining=run_mining,
                                                        run_infer=run_infer,
                                                        env_config_file_path=work_env_config_file)

        available_gpu_id: str = config.get(mir_settings.TASK_CONTEXT_KEY, {}).get('available_gpu_id', '')

        run_docker_cmd(asset_path=media_path,
                       index_file_path=work_index_file,
                       model_path=work_model_path,
                       config_file_path=work_config_file,
                       env_file_path=work_env_config_file,
                       out_path=work_out_path,
                       executor=executor,
                       executant_name=executant_name,
                       shm_size=shm_size,
                       task_type=task_id,
                       gpu_id=available_gpu_id)

        if run_infer:
            _process_infer_results(infer_result_file=os.path.join(work_out_path, 'infer-result.json'),
                                   max_boxes=_get_max_boxes(config_file))

        return MirCode.RC_OK


def _prepare_env(work_dir: str) -> Tuple[str, str, str]:
    """
    make the following dir structures:
    * work_dir
            * in
                    * assets
                    * models
            * out

    if work_dir already exists, do nothing

    Args:
        work_dir (str): work dir root
    """
    os.makedirs(os.path.join(work_dir, 'in'), exist_ok=True)
    work_assets_path = os.path.join(work_dir, 'in', 'assets')
    work_model_path = os.path.join(work_dir, 'in', 'models')
    work_out_path = os.path.join(work_dir, 'out')
    os.makedirs(work_assets_path, exist_ok=True)
    os.makedirs(work_model_path, exist_ok=True)
    os.makedirs(work_out_path, exist_ok=True)

    return work_assets_path, work_model_path, work_out_path


def _prepare_assets(index_file: str, work_index_file: str, media_path: str) -> None:
    """
    generates in container index file

    Args:
        index_file (str): path to source index file
        work_index_file (str): path to destination index file used in container
        media_path (str): path for all medias

    Raise:
        FileNotFoundError: if index_file not found, or assets in index_file not found
    """
    with open(index_file, 'r') as f:
        asset_files = f.readlines()

    # copy assets (and also generate in-container index file)
    media_keys_set = set()  # detect dumplicate media keys
    with open(work_index_file, 'w') as f:
        for src_asset_file in asset_files:
            # skip blank lines
            src_asset_file = src_asset_file.strip()
            if not src_asset_file:
                continue

            # check rel path
            if not src_asset_file.startswith('/'):
                raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                                      error_message=f"rel path not allowed: {src_asset_file}",
                                      needs_new_commit=False)

            # check repeat
            media_key = os.path.relpath(path=src_asset_file, start=media_path)
            if media_key in media_keys_set:
                raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                                      error_message=f"dumplicate image name: {media_key}, abort",
                                      needs_new_commit=False)
            media_keys_set.add(media_key)

            # write in-container index file
            f.write(f"/in/assets/{media_key}\n")

    if not media_keys_set:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                              error_message='no assets to infer, abort',
                              needs_new_commit=False)


def _process_infer_results(infer_result_file: str, max_boxes: int) -> None:
    if not os.path.isfile(infer_result_file):
        raise MirRuntimeError(error_code=MirCode.RC_CMD_NO_RESULT,
                              error_message=f"can not find result file: {infer_result_file}")

    with open(infer_result_file, 'r') as f:
        results = json.loads(f.read())

    if 'detection' in results:
        names_annotations_dict = results['detection']
        for _, annotations_dict in names_annotations_dict.items():
            if 'annotations' in annotations_dict and isinstance(annotations_dict['annotations'], list):
                annotations_dict['annotations'].sort(key=(lambda x: x['score']), reverse=True)
                annotations_dict['annotations'] = annotations_dict['annotations'][:max_boxes]

    with open(infer_result_file, 'w') as f:
        f.write(json.dumps(results, indent=4))


def _get_max_boxes(config_file: str) -> int:
    with open(config_file, 'r') as f:
        config = yaml.safe_load(f.read())

    max_boxes = config.get('max_boxes', 50)
    if not isinstance(max_boxes, int) or max_boxes <= 0:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message=f"invalid max_boxes: {max_boxes}")

    return max_boxes


# might used both by mining and infer
# public: general
def prepare_config_file(config: dict, dst_config_file: str, **kwargs: Any) -> None:
    executor_config = config[mir_settings.EXECUTOR_CONFIG_KEY]

    for k, v in kwargs.items():
        executor_config[k] = v

    logging.info(f"container config: {executor_config}")

    with open(dst_config_file, 'w') as f:
        yaml.dump(executor_config, f)


def run_docker_cmd(asset_path: str, index_file_path: str, model_path: str, config_file_path: str, env_file_path: str,
                   out_path: str, executor: str, executant_name: str, shm_size: Optional[str], task_type: str,
                   gpu_id: str) -> int:
    """ runs infer or mining docker container """
    cmd = ['nvidia-docker', 'run', '--rm']
    # path bindings
    cmd.append(f"-v{asset_path}:/in/assets:ro")
    cmd.append(f"-v{model_path}:/in/models:ro")
    cmd.append(f"-v{index_file_path}:/in/candidate-index.tsv")
    cmd.append(f"-v{config_file_path}:/in/config.yaml")
    cmd.append(f"-v{env_file_path}:/in/env.yaml")
    cmd.append(f"-v{out_path}:/out")
    # permissions and shared memory
    cmd.extend(['--user', f"{os.getuid()}:{os.getgid()}"])
    if gpu_id:
        cmd.extend(['--gpus', f"\"device={gpu_id}\""])
    if shm_size:
        cmd.append(f"--shm-size={shm_size}")
    cmd.extend(['--name', executant_name])
    cmd.append(executor)

    out_log_path = os.path.join(out_path, mir_settings.EXECUTOR_OUTLOG_NAME)
    logging.info(f"starting {task_type} docker container with cmd: {' '.join(cmd)}")
    with open(out_log_path, 'a') as f:
        # run and wait, if non-zero value returned, raise
        subprocess.run(cmd, check=True, stdout=f, stderr=f, text=True)

    return MirCode.RC_OK


# public: cli bind
def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: argparse.ArgumentParser) -> None:
    infer_arg_parser = subparsers.add_parser('infer',
                                             description='use this command to inference images',
                                             help='inference images')
    infer_arg_parser.add_argument('--index-file', dest='index_file', type=str, required=True, help='path to index file')
    infer_arg_parser.add_argument('-w',
                                  required=True,
                                  dest='work_dir',
                                  type=str,
                                  help='work place for mining and monitoring')
    infer_arg_parser.add_argument('--model-location',
                                  required=True,
                                  dest='model_location',
                                  type=str,
                                  help='model storage location for models')
    infer_arg_parser.add_argument('--model-hash',
                                  dest='model_hash',
                                  type=str,
                                  required=True,
                                  help='model hash to be used')
    infer_arg_parser.add_argument('--task-config-file',
                                  dest='config_file',
                                  type=str,
                                  required=True,
                                  help='path to executor config file')
    infer_arg_parser.add_argument('--executor',
                                  required=True,
                                  dest='executor',
                                  type=str,
                                  help="docker image name for infer or mining")
    infer_arg_parser.add_argument('--executant-name',
                                  required=False,
                                  dest='executant_name',
                                  type=str,
                                  help='docker container name for infer or mining')
    infer_arg_parser.set_defaults(func=CmdInfer)



from abc import ABC, abstractmethod
import logging
from typing import Any

from mir.tools import utils


class BaseCommand(ABC):
    """base class for all commands in mir"""

    # lifecycle
    def __init__(self, args: Any):
        self.args = args

    @utils.time_it
    def cmd_run(self) -> int:
        return self.run()

    @abstractmethod
    def run(self) -> int:
        logging.critical("BaseCommand run: this action should override in sub classes")
        return 0



import os
from typing import Any, Dict, List, Optional, Set, Tuple

from pydantic import BaseModel, validator, root_validator
import yaml

from mir.tools import utils as mir_utils


EXPECTED_FILE_VERSION = 1


class _SingleLabel(BaseModel):
    id: int
    name: str
    aliases: List[str] = []

    @validator('name')
    def _strip_and_lower_name(cls, v: str) -> str:
        return v.strip().lower()

    @validator('aliases', each_item=True)
    def _strip_and_lower_alias(cls, v: str) -> str:
        return v.strip().lower()


class _LabelStorage(BaseModel):
    version: int = EXPECTED_FILE_VERSION
    labels: List[_SingleLabel] = []
    _label_to_ids: Dict[str, Tuple[int, Optional[str]]] = {}
    _id_to_labels: Dict[int, str] = {}

    @validator('version')
    def _check_version(cls, v: int) -> int:
        if v != EXPECTED_FILE_VERSION:
            raise ValueError(f"incorrect version: {v}, needed {EXPECTED_FILE_VERSION}")
        return v

    @validator('labels')
    def _check_labels(cls, labels: List[_SingleLabel]) -> List[_SingleLabel]:
        label_names_set: Set[str] = set()
        for idx, label in enumerate(labels):
            if label.id != idx:
                raise ValueError(f"invalid label id: {label.id}, expected {idx}")

            # all label names and aliases should have no dumplicate
            name_and_aliases = label.aliases + [label.name]
            name_and_aliases_set = set(name_and_aliases)
            if len(name_and_aliases) != len(name_and_aliases_set):
                raise ValueError(f"duplicated inline label: {name_and_aliases}")
            duplicated = set.intersection(name_and_aliases_set, label_names_set)
            if duplicated:
                raise ValueError(f"duplicated: {duplicated}")
            label_names_set.update(name_and_aliases_set)
        return labels

    @root_validator
    def _generate_dicts(cls, values: dict) -> dict:
        labels: List[_SingleLabel] = values.get('labels', [])
        label_to_ids: Dict[str, Tuple[int, Optional[str]]] = {}
        id_to_labels: Dict[int, str] = {}
        for label in labels:
            _set_if_not_exists(k=label.name,
                               v=(label.id, None),
                               d=label_to_ids,
                               error_message_prefix='duplicated name')
            #   key: aliases
            for label_alias in label.aliases:
                _set_if_not_exists(k=label_alias,
                                   v=(label.id, label.name),
                                   d=label_to_ids,
                                   error_message_prefix='duplicated alias')

            # self._type_id_name_dict
            _set_if_not_exists(k=label.id,
                               v=label.name,
                               d=id_to_labels,
                               error_message_prefix='duplicated id')

        values['_label_to_ids'] = label_to_ids
        values['_id_to_labels'] = id_to_labels
        return values


def ids_file_name() -> str:
    return 'labels.yaml'


def ids_file_path(mir_root: str) -> str:
    return os.path.join(mir_utils.repo_dot_mir_path(mir_root=mir_root), ids_file_name())


def create_empty_if_not_exists(mir_root: str) -> None:
    file_path = ids_file_path(mir_root=mir_root)
    if os.path.isfile(file_path):
        return

    label_storage = _LabelStorage()
    with open(file_path, 'w') as f:
        yaml.safe_dump(label_storage.dict(), f)


class ClassIdManagerError(BaseException):
    pass


class ClassIdManager(object):
    """
    a query tool for label storage file
    """
    __slots__ = ("_storage_file_path", "_label_storage")

    # life cycle
    def __init__(self, mir_root: str) -> None:
        super().__init__()

        # it will have value iff successfully loaded
        self._storage_file_path = ''

        self.__load(ids_file_path(mir_root=mir_root))

    # private: load and unload
    def __load(self, file_path: str) -> bool:
        if not file_path:
            raise ClassIdManagerError('empty path received')
        if self._storage_file_path:
            raise ClassIdManagerError(f"already loaded from: {self._storage_file_path}")

        with open(file_path, 'r') as f:
            file_obj = yaml.safe_load(f)
        if file_obj is None:
            file_obj = {}

        self._label_storage = _LabelStorage(**file_obj)
        # save `self._storage_file_path` as a flag of successful loading
        self._storage_file_path = file_path
        return True

    # public: general
    def id_and_main_name_for_name(self, name: str) -> Tuple[int, Optional[str]]:
        """
        returns type id and main type name for main type name or alias

        Args:
            name (str): main type name or alias

        Raises:
            ClassIdManagerError: if not loaded, or name is empty, or can not find name

        Returns:
            Tuple[int, Optional[str]]: (type id, main type name)
        """
        name = name.strip().lower()
        if not self._storage_file_path:
            raise ClassIdManagerError("not loade")
        if not name:
            raise ClassIdManagerError("empty name")

        if name not in self._label_storage._label_to_ids:
            raise ClassIdManagerError(f"not exists: {name}")

        return self._label_storage._label_to_ids[name]

    def main_name_for_id(self, type_id: int) -> Optional[str]:
        """
        get main type name for type id, if not found, returns None

        Args:
            type_id (int): type id

        Returns:
            Optional[str]: corresponding main type name, if not found, returns None
        """
        return self._label_storage._id_to_labels.get(type_id, None)

    def id_for_names(self, names: List[str]) -> List[int]:
        """
        return all type ids for names

        Args:
            names (List[str]): main type names or alias

        Returns:
            List[int]: corresponding type ids
        """
        return [self.id_and_main_name_for_name(name=name)[0] for name in names]

    def all_main_names(self) -> List[str]:
        """
        Returns:
            List[str]: all main names, if not loaded, returns empty list
        """
        return list(self._label_storage._id_to_labels.values())

    def all_ids(self) -> List[int]:
        """
        Returns:
            List[int]: all class_ids, if not loaded, returns empty list
        """
        return list(self._label_storage._id_to_labels.keys())

    def size(self) -> int:
        """
        Returns:
            int: size of all type ids and main names, if not loaded, returns 0
        """
        return len(self._label_storage._id_to_labels)

    def has_name(self, name: str) -> bool:
        return name.strip().lower() in self._label_storage._label_to_ids

    def has_id(self, type_id: int) -> bool:
        return type_id in self._label_storage._id_to_labels


def _set_if_not_exists(k: Any, v: Any, d: dict, error_message_prefix: str) -> None:
    if k in d:
        raise ClassIdManagerError(f"{error_message_prefix}: {k}")
    d[k] = v



import argparse
import json
import logging
import os
from subprocess import CalledProcessError
from typing import Dict, Optional, Set

from google.protobuf import json_format
import yaml

from mir.commands import base, infer
from mir.protos import mir_command_pb2 as mirpb
from mir.tools import checker, class_ids, data_exporter, mir_storage_ops, revs_parser, utils as mir_utils
from mir.tools.code import MirCode
from mir.tools.command_run_in_out import command_run_in_out
from mir.tools.errors import MirContainerError, MirRuntimeError


class CmdMining(base.BaseCommand):
    """
    mining command

    About path mappings:
        a. work_dir/in/assets or cache -> /in/assets
        b: work_dir/in/models -> /in/models
        c: work_dir/out -> out
        d: work_dir/in/candidate-index.tsv -> /in/candidate-index.tsv
        e: work_dir/in/config.yaml -> /in/config.yaml
    """
    def run(self) -> int:
        logging.debug(f"command mining: {self.args}")

        return CmdMining.run_with_args(work_dir=self.args.work_dir,
                                       asset_cache_dir=self.args.asset_cache_dir,
                                       src_revs=self.args.src_revs,
                                       dst_rev=self.args.dst_rev,
                                       mir_root=self.args.mir_root,
                                       model_hash=self.args.model_hash,
                                       model_location=self.args.model_location,
                                       media_location=self.args.media_location,
                                       config_file=self.args.config_file,
                                       topk=self.args.topk,
                                       add_annotations=self.args.add_annotations,
                                       executor=self.args.executor,
                                       executant_name=self.args.executant_name)

    @staticmethod
    @command_run_in_out
    def run_with_args(work_dir: str,
                      asset_cache_dir: Optional[str],
                      src_revs: str,
                      dst_rev: str,
                      mir_root: str,
                      model_hash: str,
                      media_location: str,
                      model_location: str,
                      config_file: str,
                      executor: str,
                      executant_name: str,
                      topk: int = None,
                      add_annotations: bool = False) -> int:
        """
        runs a mining task \n
        Args:
            work_dir: mining docker container's work directory
            asset_cache_dir: media cache directory
            src_revs: data branch name and base task id
            dst_rev: destination branch name and task id
            mir_root: mir repo path, in order to run in non-mir folder.
            model_hash: used to target model, use prep_tid if non-set
            media_location, model_location: location of assets.
            config_file: path to the config file
            executor: executor name, currently, the docker image name
            executant_name: docker container name
            topk: top k assets you want to select in the result workspace, positive integer or None (no mining)
            add_annotations: if true, write new annotations into annotations.mir
        Returns:
            error code
        """
        if not mir_root:
            mir_root = '.'
        if not work_dir:
            logging.error('empty work_dir: abort')
            return MirCode.RC_CMD_INVALID_ARGS
        if not media_location or not model_location:
            logging.error('media or model location cannot be none!')
            return MirCode.RC_CMD_INVALID_ARGS
        if not model_hash:
            logging.error('model_hash is required.')
            return MirCode.RC_CMD_INVALID_ARGS

        src_typ_rev_tid = revs_parser.parse_single_arg_rev(src_revs, need_tid=False)
        dst_typ_rev_tid = revs_parser.parse_single_arg_rev(dst_rev, need_tid=True)

        if not config_file:
            logging.warning('empty --task-config-file, abort')
            return MirCode.RC_CMD_INVALID_ARGS
        if not os.path.isfile(config_file):
            logging.error(f"invalid --task-config-file {config_file}, not a file, abort")
            return MirCode.RC_CMD_INVALID_ARGS

        if not executor:
            logging.error('empty --executor, abort')
            return MirCode.RC_CMD_INVALID_ARGS

        return_code = checker.check(mir_root, [checker.Prerequisites.IS_INSIDE_MIR_REPO])
        if return_code != MirCode.RC_OK:
            return return_code

        # check `topk` and `add_annotations`
        mir_metadatas: mirpb.MirMetadatas = mir_storage_ops.MirStorageOps.load_single_storage(
            mir_root=mir_root,
            mir_branch=src_typ_rev_tid.rev,
            mir_task_id=src_typ_rev_tid.tid,
            ms=mirpb.MirStorage.MIR_METADATAS)
        assets_count = len(mir_metadatas.attributes)
        if assets_count == 0:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_MERGE_ERROR,
                                  error_message='no assets found in metadatas.mir')
        if topk:
            if topk >= assets_count:
                logging.warning(f"topk: {topk} >= assets count: {assets_count}, skip mining")
                topk = None
            elif topk <= 0:
                logging.error(f"invalid --topk: {topk}")
                return MirCode.RC_CMD_INVALID_ARGS
        # topk can be None (means no mining), or in interval (0, assets_count) (means mining and select topk)

        work_in_path = os.path.join(work_dir, 'in')  # docker container's input data directory
        work_asset_path = asset_cache_dir or os.path.join(work_in_path, 'assets')
        work_model_path = os.path.join(work_in_path, 'models')
        work_index_file = os.path.join(work_in_path, 'candidate-src-index.tsv')
        work_out_path = os.path.join(work_dir, 'out')  # docker container's output data directory

        ret = _prepare_env(export_root=work_dir,
                           work_in_path=work_in_path,
                           work_out_path=work_out_path,
                           work_asset_path=work_asset_path,
                           work_model_path=work_model_path)
        if ret != MirCode.RC_OK:
            logging.error('mining enviroment prepare error!')
            return ret

        _prepare_assets(mir_metadatas=mir_metadatas,
                        mir_root=mir_root,
                        src_rev_tid=src_typ_rev_tid,
                        media_location=media_location,
                        work_asset_path=work_asset_path,
                        work_index_file=work_index_file)

        return_code = MirCode.RC_OK
        return_msg = ''
        try:
            infer.CmdInfer.run_with_args(work_dir=work_dir,
                                         media_path=work_asset_path,
                                         model_location=model_location,
                                         model_hash=model_hash,
                                         index_file=work_index_file,
                                         config_file=config_file,
                                         task_id=dst_typ_rev_tid.tid,
                                         shm_size=_get_shm_size(config_file),
                                         executor=executor,
                                         executant_name=executant_name,
                                         run_infer=add_annotations,
                                         run_mining=(topk is not None))
        except CalledProcessError:
            return_code = MirCode.RC_CMD_CONTAINER_ERROR
            return_msg = mir_utils.collect_executor_outlog_tail(work_dir=work_dir)
        # catch other exceptions in command_run_in_out

        task = mir_storage_ops.create_task(task_type=mirpb.TaskTypeMining,
                                           task_id=dst_typ_rev_tid.tid,
                                           message='mining',
                                           model_hash=model_hash,
                                           src_revs=src_typ_rev_tid.rev_tid,
                                           dst_rev=dst_typ_rev_tid.rev_tid,
                                           return_code=return_code,
                                           return_msg=return_msg,
                                           executor=executor)
        if return_code != MirCode.RC_OK:
            raise MirContainerError(error_message='mining container error occured', task=task)

        _process_results(mir_root=mir_root,
                         export_out=work_out_path,
                         dst_typ_rev_tid=dst_typ_rev_tid,
                         src_typ_rev_tid=src_typ_rev_tid,
                         topk=topk,
                         add_annotations=add_annotations,
                         task=task)
        logging.info(f"mining done, results at: {work_out_path}")

        return MirCode.RC_OK


# protected: post process
def _process_results(mir_root: str, export_out: str, dst_typ_rev_tid: revs_parser.TypRevTid,
                     src_typ_rev_tid: revs_parser.TypRevTid, topk: Optional[int], add_annotations: bool,
                     task: mirpb.Task) -> int:
    # step 1: build topk results:
    #   read old
    [mir_metadatas, mir_annotations] = mir_storage_ops.MirStorageOps.load_multiple_storages(
        mir_root=mir_root,
        mir_branch=src_typ_rev_tid.rev,
        mir_task_id=src_typ_rev_tid.tid,
        ms_list=[mirpb.MirStorage.MIR_METADATAS, mirpb.MirStorage.MIR_ANNOTATIONS],
        as_dict=False,
    )

    #   parse new: topk and new annotations (both optional)
    topk_result_file_path = os.path.join(export_out, 'result.tsv')
    asset_ids_set = (_get_topk_asset_ids(file_path=topk_result_file_path, topk=topk)
                     if topk is not None else set(mir_metadatas.attributes.keys()))

    infer_result_file_path = os.path.join(export_out, 'infer-result.json')
    cls_id_mgr = class_ids.ClassIdManager(mir_root=mir_root)
    asset_id_to_annotations = (_get_infer_annotations(file_path=infer_result_file_path,
                                                      asset_ids_set=asset_ids_set,
                                                      cls_id_mgr=cls_id_mgr) if add_annotations else {})

    # step 2: update mir data files
    #   update mir metadatas
    matched_mir_metadatas = mirpb.MirMetadatas()
    for asset_id in asset_ids_set:
        matched_mir_metadatas.attributes[asset_id].CopyFrom(mir_metadatas.attributes[asset_id])
    logging.info(f"matched: {len(matched_mir_metadatas.attributes)}, overriding metadatas.mir")

    #   update mir annotations
    matched_mir_annotations = mirpb.MirAnnotations()
    matched_task_annotation = matched_mir_annotations.task_annotations[dst_typ_rev_tid.tid]
    if add_annotations:
        # add new
        for asset_id, single_image_annotations in asset_id_to_annotations.items():
            matched_task_annotation.image_annotations[asset_id].CopyFrom(single_image_annotations)
    else:
        # use old
        task_annotation = mir_annotations.task_annotations[mir_annotations.head_task_id]
        joint_asset_ids_set = set(task_annotation.image_annotations.keys()) & asset_ids_set
        for asset_id in joint_asset_ids_set:
            matched_task_annotation.image_annotations[asset_id].CopyFrom(task_annotation.image_annotations[asset_id])

    #   mir_keywords: auto generated from mir_annotations, so do nothing

    # step 3: store results and commit.
    mir_datas = {
        mirpb.MirStorage.MIR_METADATAS: matched_mir_metadatas,
        mirpb.MirStorage.MIR_ANNOTATIONS: matched_mir_annotations,
    }
    mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
                                                  his_branch=src_typ_rev_tid.rev,
                                                  mir_branch=dst_typ_rev_tid.rev,
                                                  mir_datas=mir_datas,
                                                  task=task)

    return MirCode.RC_OK


def _get_topk_asset_ids(file_path: str, topk: int) -> Set[str]:
    if not os.path.isfile(file_path):
        raise MirRuntimeError(MirCode.RC_CMD_NO_RESULT, f"Cannot find result file {file_path}")

    asset_ids_set: Set[str] = set()
    idx_cnt = 0
    with open(file_path) as ret_f:
        for line in ret_f:
            line = line.strip()
            if not line:
                continue
            if idx_cnt >= topk:
                break
            asset_ids_set.add(os.path.splitext(os.path.basename(line.split('\t')[0]))[0])  # main file name (asset id)
            idx_cnt += 1
    logging.info(f"top {len(asset_ids_set)} samples found")
    return asset_ids_set


def _get_infer_annotations(file_path: str, asset_ids_set: Set[str],
                           cls_id_mgr: class_ids.ClassIdManager) -> Dict[str, mirpb.SingleImageAnnotations]:
    asset_id_to_annotations: dict = {}
    with open(file_path, 'r') as f:
        results = json.loads(f.read())

    if 'detection' not in results or not isinstance(results['detection'], dict):
        logging.error('invalid infer-result.json')
        return asset_id_to_annotations

    names_annotations_dict = results['detection']
    for asset_name, annotations_dict in names_annotations_dict.items():
        if 'annotations' not in annotations_dict or not isinstance(annotations_dict['annotations'], list):
            continue
        asset_id = os.path.splitext(os.path.basename(asset_name))[0]
        if asset_id not in asset_ids_set:
            logging.debug(f"unknown asset name: {asset_name}, ignore")
            continue
        single_image_annotations = mirpb.SingleImageAnnotations()
        for idx, annotation_dict in enumerate(annotations_dict['annotations']):
            annotation = mirpb.Annotation()
            annotation.index = idx
            json_format.ParseDict(annotation_dict['box'], annotation.box)
            annotation.class_id = cls_id_mgr.id_and_main_name_for_name(annotation_dict['class_name'])[0]
            annotation.score = float(annotation_dict.get('score', 0))
            single_image_annotations.annotations.append(annotation)
        asset_id_to_annotations[asset_id] = single_image_annotations
    return asset_id_to_annotations


# protected: pre process
def _prepare_env(export_root: str, work_in_path: str, work_out_path: str, work_asset_path: str,
                 work_model_path: str) -> int:
    os.makedirs(export_root, exist_ok=True)
    os.makedirs(work_in_path, exist_ok=True)
    os.makedirs(work_out_path, exist_ok=True)
    os.makedirs(work_asset_path, exist_ok=True)
    os.makedirs(work_model_path, exist_ok=True)

    return MirCode.RC_OK


def _prepare_assets(mir_metadatas: mirpb.MirMetadatas, mir_root: str, src_rev_tid: revs_parser.TypRevTid,
                    media_location: str, work_asset_path: str, work_index_file: str) -> None:
    img_list = set(mir_metadatas.attributes.keys())
    data_exporter.export(mir_root=mir_root,
                         assets_location=media_location,
                         class_type_ids={},
                         asset_ids=img_list,
                         asset_dir=work_asset_path,
                         annotation_dir='',
                         need_ext=True,
                         need_id_sub_folder=True,
                         base_branch=src_rev_tid.rev,
                         base_task_id=src_rev_tid.tid,
                         format_type=data_exporter.ExportFormat.EXPORT_FORMAT_NO_ANNOTATION,
                         index_file_path=work_index_file,
                         index_assets_prefix=work_asset_path)


def _get_shm_size(mining_config_file_path: str) -> str:
    with open(mining_config_file_path, 'r') as f:
        mining_config = yaml.safe_load(f.read())
    if 'shm_size' not in mining_config:
        return '16G'
    return mining_config['shm_size']


# public: arg parser
def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: argparse.ArgumentParser) -> None:
    mining_arg_parser = subparsers.add_parser('mining',
                                              parents=[parent_parser],
                                              description='use this command to mine in current workspace',
                                              help='mine current workspace')
    mining_arg_parser.add_argument('-w',
                                   required=True,
                                   dest='work_dir',
                                   type=str,
                                   help='work place for mining and monitoring')
    mining_arg_parser.add_argument('--asset-cache-dir',
                                   required=False,
                                   dest='asset_cache_dir',
                                   type=str,
                                   help='media cache directory')
    mining_arg_parser.add_argument('--model-location',
                                   required=True,
                                   dest='model_location',
                                   type=str,
                                   help='model storage location')
    mining_arg_parser.add_argument('--media-location',
                                   required=True,
                                   dest='media_location',
                                   type=str,
                                   help='media storage location')
    mining_arg_parser.add_argument('--topk',
                                   dest='topk',
                                   type=int,
                                   required=False,
                                   help='if set, discard samples out of topk, sorting by scores.')
    mining_arg_parser.add_argument('--add-annotations',
                                   dest='add_annotations',
                                   action='store_true',
                                   required=False,
                                   help='if set, also add inference result to annotations')
    mining_arg_parser.add_argument('--model-hash',
                                   dest='model_hash',
                                   type=str,
                                   required=True,
                                   help='model hash to be used')
    mining_arg_parser.add_argument('--src-revs',
                                   dest='src_revs',
                                   type=str,
                                   required=True,
                                   help='rev@bid: source rev and base task id')
    mining_arg_parser.add_argument('--dst-rev',
                                   dest='dst_rev',
                                   type=str,
                                   required=True,
                                   help='rev@tid: destination branch name and task id')
    mining_arg_parser.add_argument('--task-config-file',
                                   dest='config_file',
                                   type=str,
                                   required=True,
                                   help='path to executor config file')
    mining_arg_parser.add_argument('--executor',
                                   required=True,
                                   dest='executor',
                                   type=str,
                                   help='docker image name for mining')
    mining_arg_parser.add_argument('--executant-name',
                                   required=False,
                                   dest='executant_name',
                                   type=str,
                                   help='docker container name for mining')
    mining_arg_parser.set_defaults(func=CmdMining)



import os
from typing import List

import yaml

from mir.tools import class_ids
from mir.tools import utils as mir_utils


def context_file_path_from_mir_root(mir_root: str) -> str:
    return os.path.join(mir_utils.repo_dot_mir_path(mir_root=mir_root), 'context.yaml')


# save and load
def load(mir_root: str) -> List[int]:
    context_file_path = context_file_path_from_mir_root(mir_root)
    if not os.path.isfile(context_file_path):
        return []

    with open(context_file_path, 'r') as f:
        context_obj = yaml.safe_load(f)
        return context_obj.get('project', {}).get('class_ids', [])


def save(mir_root: str, project_class_ids: List[int]) -> None:
    context_file_path = context_file_path_from_mir_root(mir_root)

    with open(context_file_path, 'w') as f:
        yaml.safe_dump({'project': {'class_ids': project_class_ids}}, f)


# general
def check_class_names(mir_root: str, current_class_names: List[str]) -> bool:
    """
    check `current_class_names` matches mir repo's project class ids settings

    if mir repo has project class ids settings, this function returns True if they are equal

    if mir repo has no project class ids settings, this function always returns True, meaning they are always matched
    """
    class_id_manager = class_ids.ClassIdManager(mir_root)
    current_class_ids = class_id_manager.id_for_names(current_class_names)
    return check_class_ids(mir_root=mir_root, current_class_ids=current_class_ids)


def check_class_ids(mir_root: str, current_class_ids: List[int]) -> bool:
    """
    check `current_class_ids` matches mir repo's project class ids settings

    if mir repo has project class ids settings, this function returns True if they are equal

    if mir repo has no project class ids settings, this function always returns True, meaning they are always matched
    """
    project_class_ids = load(mir_root=mir_root)
    if not project_class_ids:
        # if this mir repo not binded to project, treat as equal
        return True
    user_class_ids = class_ids.ClassIdManager(mir_root).all_ids()
    return set(current_class_ids) <= set(user_class_ids)



from typing import Any

from mir.tools.code import MirCode


class MirError(Exception):
    """
    error raised by ymir-command
    """
    def __init__(self, error_code: int, error_message: str) -> None:
        super().__init__()
        self.error_code = error_code
        self.error_message = error_message

    def __str__(self) -> str:
        return f"code: {self.error_code}, content: {self.error_message}"


class MirRuntimeError(MirError):
    """

    error raised by ymir-command

    if raised from a function surrounded by decorator `@command_run_in_out`,
    new commit will made if `needs_new_commit` true

    """
    def __init__(self, error_code: int, error_message: str, needs_new_commit: bool = True, task: Any = None):
        super().__init__(error_code=error_code, error_message=error_message)
        self.needs_new_commit = needs_new_commit
        self.task = task

    def __copy__(self) -> 'MirRuntimeError':
        return MirRuntimeError(error_code=self.error_code,
                               error_message=self.error_message,
                               needs_new_commit=self.needs_new_commit,
                               task=self.task)

    def __str__(self) -> str:
        return f"code: {self.error_code}, message: {self.error_message}"


class MirContainerError(MirRuntimeError):
    def __init__(self, error_message: str, task: Any):
        super().__init__(error_code=MirCode.RC_CMD_CONTAINER_ERROR,
                         error_message=error_message,
                         needs_new_commit=True,
                         task=task)



import enum
import logging
import os
import sys
from typing import Callable, List

from mir.tools import class_ids, mir_repo_utils
from mir.tools.code import MirCode


@enum.unique
class Prerequisites(enum.IntEnum):
    NOTHING = 0
    IS_INSIDE_GIT_REPO = enum.auto()
    IS_OUTSIDE_GIT_REPO = enum.auto()
    IS_INSIDE_MIR_REPO = enum.auto()
    IS_OUTSIDE_MIR_REPO = enum.auto()
    IS_DIRTY = enum.auto()
    IS_CLEAN = enum.auto()
    HAVE_LABELS = enum.auto()
    HAVE_NO_LABELS = enum.auto()


_DEFAULT_PREREQUISTITES = [Prerequisites.IS_INSIDE_MIR_REPO, Prerequisites.IS_CLEAN, Prerequisites.HAVE_LABELS]


# error messages for check failed (if prerequisites not satisfied)
_ERROR_INFOS = {
    Prerequisites.NOTHING: '',
    Prerequisites.IS_INSIDE_GIT_REPO: 'mir_root is not a git repo',
    Prerequisites.IS_OUTSIDE_GIT_REPO: 'mir_root is already a git repo',
    Prerequisites.IS_INSIDE_MIR_REPO: 'mir_root is not a mir repo',
    Prerequisites.IS_OUTSIDE_MIR_REPO: 'mir_root is already a mir repo',
    Prerequisites.IS_DIRTY: 'mir repo is clean (need dirty)',
    Prerequisites.IS_CLEAN: 'mir repo is dirty (need clean)',
    Prerequisites.HAVE_LABELS: f"can not find {class_ids.ids_file_name()}",
    Prerequisites.HAVE_NO_LABELS: f"already have {class_ids.ids_file_name()}",
}


# check mir root
def check(mir_root: str, prerequisites: List[Prerequisites] = _DEFAULT_PREREQUISTITES) -> int:
    if not mir_root.startswith('/'):
        logging.warning(f"check failed: invalid mir_root: {mir_root or '(empty)'}, needs absolute path")
        return MirCode.RC_CMD_INVALID_ARGS

    for item in prerequisites:
        checker_func: Callable = getattr(sys.modules[__name__], f"_check_{item.name.lower()}")
        return_code = checker_func(mir_root)
        if return_code != MirCode.RC_OK:
            logging.error(f"check failed: {_ERROR_INFOS[item]}")
            return return_code
    return MirCode.RC_OK


def _check_nothing(mir_root: str) -> int:
    return MirCode.RC_OK


def _check_is_inside_git_repo(mir_root: str) -> int:
    return (MirCode.RC_OK if os.path.isdir(os.path.join(mir_root, ".git")) else MirCode.RC_CMD_INVALID_ARGS)


def _check_is_outside_git_repo(mir_root: str) -> int:
    return (MirCode.RC_OK if not os.path.isdir(os.path.join(mir_root, ".git")) else MirCode.RC_CMD_INVALID_ARGS)


def _check_is_inside_mir_repo(mir_root: str) -> int:
    return (MirCode.RC_OK if os.path.isdir(os.path.join(mir_root, ".git"))
            and os.path.isdir(os.path.join(mir_root, ".mir")) else MirCode.RC_CMD_INVALID_MIR_REPO)


def _check_is_outside_mir_repo(mir_root: str) -> int:
    return (MirCode.RC_OK if not os.path.isdir(os.path.join(mir_root, ".git"))
            or not os.path.isdir(os.path.join(mir_root, ".mir")) else MirCode.RC_CMD_INVALID_ARGS)


def _check_is_dirty(mir_root: str) -> int:
    is_dirty = mir_repo_utils.mir_check_repo_git_dirty(mir_root)
    return MirCode.RC_OK if is_dirty else MirCode.RC_CMD_INVALID_ARGS


def _check_is_clean(mir_root: str) -> int:
    is_dirty = mir_repo_utils.mir_check_repo_git_dirty(mir_root)
    return MirCode.RC_OK if not is_dirty else MirCode.RC_CMD_DIRTY_REPO


def _check_have_labels(mir_root: str) -> int:
    have_labels = os.path.isfile(class_ids.ids_file_path(mir_root))
    return MirCode.RC_OK if have_labels else MirCode.RC_CMD_INVALID_MIR_REPO


def _check_have_no_labels(mir_root: str) -> int:
    have_labels = os.path.isfile(class_ids.ids_file_path(mir_root))
    return MirCode.RC_OK if not have_labels else MirCode.RC_CMD_INVALID_MIR_REPO



import argparse
import logging

from mir.commands import base
from mir.tools import checker, mir_repo_utils
from mir.tools.code import MirCode


class CmdStatus(base.BaseCommand):
    def run(self) -> int:
        logging.debug("command status: %s", self.args)

        return self.run_with_args(mir_root=self.args.mir_root)

    @staticmethod
    def run_with_args(mir_root: str) -> int:
        return_code = checker.check(mir_root, [checker.Prerequisites.IS_INSIDE_MIR_REPO])
        if return_code != MirCode.RC_OK:
            return return_code

        repo_dirty = mir_repo_utils.mir_check_repo_git_dirty(mir_root=mir_root)
        logging.info('repo: dirty' if repo_dirty else 'repo: clean')

        return MirCode.RC_OK


def bind_to_subparsers(subparsers: argparse._SubParsersAction,
                       parent_parser: argparse.ArgumentParser) -> None:
    status_arg_parser = subparsers.add_parser("status",
                                              parents=[parent_parser],
                                              description="use this command to show current workspace status",
                                              help="show current workspace status")
    status_arg_parser.set_defaults(func=CmdStatus)



import argparse
import logging
from typing import Any, Callable, List, Tuple, Optional, Set, Union

from mir.commands import base
from mir.protos import mir_command_pb2 as mirpb
from mir.tools import checker, class_ids, mir_repo_utils, mir_storage, mir_storage_ops, revs_parser
from mir.tools.code import MirCode
from mir.tools.command_run_in_out import command_run_in_out
from mir.tools.errors import MirRuntimeError
from mir.tools.phase_logger import PhaseLoggerCenter

# type for function `__include_match` and `__exclude_match`
__IncludeExcludeCallableType = Callable[[Set[str], mirpb.MirKeywords, str, Any], Set[str]]


class CmdFilter(base.BaseCommand):
    # private: misc
    @staticmethod
    def __preds_set_from_str(preds_str: str, cls_mgr: class_ids.ClassIdManager) -> Set[int]:
        if not preds_str:
            return set()

        return set(cls_mgr.id_for_names(preds_str.split(";")))

    @staticmethod
    def __include_match(asset_ids_set: Set[str], mir_keywords: mirpb.MirKeywords, attr_name: str,
                        in_set: Set[Union[int, str]]) -> Set[str]:
        # if don't need include match, returns all
        if not in_set:
            return asset_ids_set

        matched_asset_ids_set = set()  # type: Set[str]
        for asset_id in asset_ids_set:
            if asset_id not in mir_keywords.keywords:
                continue

            keyids_set = set(getattr(mir_keywords.keywords[asset_id], attr_name))
            if not keyids_set:
                continue

            if keyids_set & in_set:
                matched_asset_ids_set.add(asset_id)
        return matched_asset_ids_set

    @staticmethod
    def __exclude_match(asset_ids_set: Set[str], mir_keywords: mirpb.MirKeywords, attr_name: str,
                        ex_set: Set[Union[int, str]]) -> Set[str]:
        # if don't need excludes filter, returns all
        if not ex_set:
            return asset_ids_set

        matched_asset_ids_set = set()  # type: Set[str]
        for asset_id in asset_ids_set:
            if asset_id in mir_keywords.keywords:
                keyids_set = set(getattr(mir_keywords.keywords[asset_id], attr_name))
                if keyids_set & ex_set:
                    continue
            matched_asset_ids_set.add(asset_id)
        return matched_asset_ids_set

    # public: run cmd
    @staticmethod
    @command_run_in_out
    def run_with_args(mir_root: str, in_cis: Optional[str], ex_cis: Optional[str], in_cks: Optional[str],
                      ex_cks: Optional[str], src_revs: str, dst_rev: str, work_dir: str) -> int:  # type: ignore
        # check args
        in_cis = in_cis.strip().lower() if in_cis else ''
        ex_cis = ex_cis.strip().lower() if ex_cis else ''
        in_cks = in_cks.strip() if in_cks else ''
        ex_cks = ex_cks.strip() if ex_cks else ''

        src_typ_rev_tid = revs_parser.parse_single_arg_rev(src_revs, need_tid=False)
        dst_typ_rev_tid = revs_parser.parse_single_arg_rev(dst_rev, need_tid=True)

        PhaseLoggerCenter.create_phase_loggers(top_phase='filter',
                                               monitor_file=mir_repo_utils.work_dir_to_monitor_file(work_dir),
                                               task_name=dst_typ_rev_tid.tid)

        return_code = checker.check(mir_root,
                                    [checker.Prerequisites.IS_INSIDE_MIR_REPO, checker.Prerequisites.HAVE_LABELS])
        if return_code != MirCode.RC_OK:
            return return_code

        [mir_metadatas, mir_annotations, mir_keywords, mir_tasks,
         _] = mir_storage_ops.MirStorageOps.load_multiple_storages(mir_root=mir_root,
                                                                   mir_branch=src_typ_rev_tid.rev,
                                                                   mir_task_id=src_typ_rev_tid.tid,
                                                                   ms_list=mir_storage.get_all_mir_storage(),
                                                                   as_dict=False)
        task_id = dst_typ_rev_tid.tid
        base_task_id = mir_annotations.head_task_id

        PhaseLoggerCenter.update_phase(phase='filter.read')

        if task_id in mir_tasks.tasks:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_BRANCH_OR_TAG,
                                  error_message=f"invalid args: task id already exists: {task_id}")
        if not base_task_id:
            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_MIR_REPO,
                                  error_message='no base task id in tasks.mir')

        assert len(mir_annotations.task_annotations.keys()) == 1
        base_task_annotations = mir_annotations.task_annotations[base_task_id]  # type: mirpb.SingleTaskAnnotations

        class_manager = class_ids.ClassIdManager(mir_root=mir_root)
        preds_set = CmdFilter.__preds_set_from_str(in_cis, class_manager)  # type: Set[int]
        excludes_set = CmdFilter.__preds_set_from_str(ex_cis, class_manager)  # type: Set[int]
        ck_preds_set = {ck.strip() for ck in in_cks.split(";")} if in_cks else set()
        ck_excludes_set = {ck.strip() for ck in ex_cks.split(";")} if ex_cks else set()

        asset_ids_set = set(mir_metadatas.attributes.keys())
        match_functions: List[Tuple[__IncludeExcludeCallableType, Union[Set[str], Set[int]], str, str]] = [
            (CmdFilter.__include_match, preds_set, 'predifined_keyids', 'select cis'),
            (CmdFilter.__exclude_match, excludes_set, 'predifined_keyids', 'exclude cis'),
            (CmdFilter.__include_match, ck_preds_set, 'customized_keywords', 'select cks'),
            (CmdFilter.__exclude_match, ck_excludes_set, 'customized_keywords', 'exclude cks')
        ]
        for match_func, ci_ck_conditions, attr_name, message in match_functions:
            if ci_ck_conditions:
                logging.info(f"assets count before {message}: {len(asset_ids_set)}")
                asset_ids_set = match_func(asset_ids_set, mir_keywords, attr_name, ci_ck_conditions)
                logging.info(f"assets count after {message}: {len(asset_ids_set)}")

        matched_mir_metadatas = mirpb.MirMetadatas()
        matched_mir_annotations = mirpb.MirAnnotations()
        matched_mir_keywords = mirpb.MirKeywords()

        # generate matched metadatas, annotations and keywords
        for asset_id in asset_ids_set:
            # generate `matched_mir_metadatas`
            asset_attr = mir_metadatas.attributes[asset_id]
            matched_mir_metadatas.attributes[asset_id].CopyFrom(asset_attr)

        joint_ids = asset_ids_set & mir_keywords.keywords.keys()
        for asset_id in joint_ids:
            # generate `matched_mir_keywords`
            matched_mir_keywords.keywords[asset_id].CopyFrom(mir_keywords.keywords[asset_id])

        # generate `matched_mir_annotations`
        joint_ids = asset_ids_set & base_task_annotations.image_annotations.keys()
        for asset_id in joint_ids:
            matched_mir_annotations.task_annotations[task_id].image_annotations[asset_id].CopyFrom(
                base_task_annotations.image_annotations[asset_id])

        logging.info("matched: %d, overriding current mir repo", len(matched_mir_metadatas.attributes))

        PhaseLoggerCenter.update_phase(phase='filter.change')

        commit_message = f"filter select: {in_cis} exclude: {ex_cis}"
        task = mir_storage_ops.create_task(task_type=mirpb.TaskType.TaskTypeFilter,
                                           task_id=task_id,
                                           message=commit_message,
                                           src_revs=src_revs,
                                           dst_rev=dst_rev)
        matched_mir_contents = {
            mirpb.MirStorage.MIR_METADATAS: matched_mir_metadatas,
            mirpb.MirStorage.MIR_ANNOTATIONS: matched_mir_annotations,
        }

        mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
                                                      mir_branch=dst_typ_rev_tid.rev,
                                                      his_branch=src_typ_rev_tid.rev,
                                                      mir_datas=matched_mir_contents,
                                                      task=task)

        return MirCode.RC_OK

    def run(self) -> int:
        logging.debug("command filter: %s", self.args)
        return CmdFilter.run_with_args(mir_root=self.args.mir_root,
                                       in_cis=self.args.in_cis,
                                       ex_cis=self.args.ex_cis,
                                       in_cks=self.args.in_cks,
                                       ex_cks=self.args.ex_cks,
                                       src_revs=self.args.src_revs,
                                       dst_rev=self.args.dst_rev,
                                       work_dir=self.args.work_dir)


def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: argparse.ArgumentParser) -> None:
    filter_arg_parser = subparsers.add_parser("filter",
                                              parents=[parent_parser],
                                              description="use this command to filter assets",
                                              help="filter assets")
    filter_arg_parser.add_argument("-p", dest="in_cis", type=str, help="type names")
    filter_arg_parser.add_argument("-P", dest="ex_cis", type=str, help="exclusive type names")
    filter_arg_parser.add_argument("-c", dest="in_cks", type=str, help="customized keywords")
    filter_arg_parser.add_argument("-C", dest="ex_cks", type=str, help="excludsive customized keywords")
    filter_arg_parser.add_argument("--src-revs", dest="src_revs", type=str, help="type:rev@bid")
    filter_arg_parser.add_argument("--dst-rev", dest="dst_rev", type=str, help="rev@tid")
    filter_arg_parser.add_argument('-w', dest='work_dir', type=str, required=False, help='working directory')
    filter_arg_parser.set_defaults(func=CmdFilter)



import logging
import os
import time
from PIL import Image, ImageFile, UnidentifiedImageError
from typing import Tuple

from mir.tools.code import MirCode
from mir.tools.errors import MirRuntimeError
from mir.tools.phase_logger import PhaseLoggerCenter
from mir.protos import mir_command_pb2 as mirpb

ImageFile.LOAD_TRUNCATED_IMAGES = True


def _generate_metadata_mir_pb(mir_metadatas: mirpb.MirMetadatas, dataset_name: str, sha1s: list, hashed_asset_root: str,
                              phase: str) -> int:
    """
    generate mirpb.MirMetadatas from sha1s
    """
    current_timestamp = int(time.time())  # this is a fake timestamp
    timestamp = mirpb.Timestamp()
    timestamp.start = current_timestamp
    timestamp.duration = 0  # image has no duraton

    unknown_format_count = 0

    sha1s_count = len(sha1s)
    for idx, val in enumerate(sha1s):
        metadata_attributes = mirpb.MetadataAttributes()
        metadata_attributes.timestamp.CopyFrom(timestamp)
        metadata_attributes.dataset_name = dataset_name

        # read file
        # if any exception occured, exit without any handler
        hashed_asset_path = os.path.join(hashed_asset_root, val)
        asset_type, width, height, channel = _type_shape_for_asset(hashed_asset_path)
        if asset_type == mirpb.AssetTypeUnknown:
            logging.warning(f"ignore asset with unknown format, id: {val}")
            unknown_format_count += 1
            continue
        metadata_attributes.asset_type = asset_type
        metadata_attributes.width = width
        metadata_attributes.height = height
        metadata_attributes.image_channels = channel

        mir_metadatas.attributes[val].CopyFrom(metadata_attributes)

        if idx > 0 and idx % 5000 == 0:
            PhaseLoggerCenter.update_phase(phase=phase, local_percent=(idx / sha1s_count))

    if unknown_format_count > 0:
        logging.warning(f"unknown format asset count: {unknown_format_count}")

    return MirCode.RC_OK


_ASSET_TYPE_STR_TO_ENUM_MAPPING = {
    'jpeg': mirpb.AssetTypeImageJpeg,
    'jpg': mirpb.AssetTypeImageJpeg,
    'png': mirpb.AssetTypeImagePng,
    'bmp': mirpb.AssetTypeImageBmp,
}


def _type_shape_for_asset(asset_path: str) -> Tuple['mirpb.AssetType.V', int, int, int]:
    if not asset_path:
        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                              error_message='_type_shape_for_asset: empty asset_path')

    try:
        asset_image = Image.open(asset_path)
        asset_type_str: str = asset_image.format.lower()  # type: ignore
    except UnidentifiedImageError as e:
        logging.info(f"{type(e).__name__}: {e} asset_path: {asset_path}")
        asset_type_str = ''  # didn't set it to 'unknown' as what i did in utils.py, because this is easy to compare

    if asset_type_str in _ASSET_TYPE_STR_TO_ENUM_MAPPING:
        width, height = asset_image.size
        channel = len(asset_image.getbands())
        return (_ASSET_TYPE_STR_TO_ENUM_MAPPING[asset_type_str], width, height, channel)
    else:
        return (mirpb.AssetTypeUnknown, 0, 0, 0)


def import_metadatas(mir_metadatas: mirpb.MirMetadatas,
                     dataset_name: str,
                     in_sha1_path: str,
                     hashed_asset_root: str,
                     phase: str = '') -> int:
    # if not enough args, abort
    if (not in_sha1_path or not dataset_name or not hashed_asset_root):
        logging.error('invalid in_sha1_path, dataset_name or hashed_asset_root')
        return MirCode.RC_CMD_INVALID_ARGS

    if not mir_metadatas:
        # some errors occured, show error message
        logging.error('mir_metadatas empty')
        return MirCode.RC_CMD_INVALID_MIR_REPO

    # read sha1
    sha1s = []
    with open(in_sha1_path, "r") as in_file:
        for line in in_file.readlines():
            if not line or not line.strip():
                continue
            line_components = line.strip().split()
            if not line_components[0]:
                continue
            sha1s.append(line_components[0])
    if not sha1s:
        logging.error(f"no sha1s found in {in_sha1_path}, exit")
        return MirCode.RC_CMD_INVALID_ARGS

    # generate mir_metadatas
    ret = _generate_metadata_mir_pb(mir_metadatas=mir_metadatas,
                                    dataset_name=dataset_name,
                                    sha1s=sha1s,
                                    hashed_asset_root=hashed_asset_root,
                                    phase=phase)
    return ret



from setuptools import setup, find_packages

from mir import version

print(version.__version__)

# Module dependencies
requirements = []
with open('requirements.txt') as f:
    for line in f.read().splitlines():
        requirements.append(line)

setup(
    name='ymir-cmd',
    version=version.__version__,
    python_requires=">=3.8.10",
    author_email="contact.viesc@gmail.com",
    description="mir: A data version control tool for YMIR",
    url="https://github.com/IndustryEssentials/ymir",
    packages=find_packages(exclude=["*tests*"]),
    install_requires=requirements,
    include_package_data=True,
    entry_points={"console_scripts": ["mir = mir.main:main"]},
)
